/***************************************************************************
 *   Copyright (C) 2005 by John Bennett   *
 *   n4xi@arrl.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

//
// sun.cpp - routines to determine sunrise/sunset times for a given locale
//

#define _SUN_CPP_

#include <math.h>

#include "sun.h"
#include "wxnmath.h"
#include "tz.h"


//////////////////////////////////////////////////////////////////////////////
// CalculateRiseSet - arguments split so that nodeblk and noderx can both
// use the routine.
//

void CalculateRiseSet( noderef * pInfo, nodedata * pData )
{
    BOOL bDstFlag;

    if( pInfo->dLatitude && pInfo->dLongitude )
    {
        if( pInfo->bDst && g_Time.DstNow() )
            bDstFlag = TRUE;
        else
            bDstFlag = FALSE;
        pData->tSunrise = SunRiseSet( SUNRISE, pInfo->dLatitude, pInfo->dLongitude, pInfo->nTimeZone, bDstFlag );
        pData->tSunset = SunRiseSet( SUNSET, pInfo->dLatitude, pInfo->dLongitude, pInfo->nTimeZone, bDstFlag );
    }
    else
    {
        pData->tSunrise = 0;
        pData->tSunset = 0;
    }
}


///////////////////////////////////////
// SunNormalize - used by SunRiseSet()
//

double SunNormalize( double dValue )
{
    while( dValue < 0 )
        dValue += QUAD4;

    while( dValue >= QUAD4 )
        dValue -= QUAD4;

    return dValue;
}


/////////////////////////////////////////////////////////////////////
//  SunRiseSet - determine sunrise or sunset time. Return number of
//  seconds since midnight for occurence.
//

time_t SunRiseSet( int nRiseSet, double dStnLat, double dStnLong, int nStnZone, BOOL bDST )
{
    double dMeanAnomaly, dSolarLong;
    double q, s, t;
    double dApproxTime, dAscension, dDeclination;
    double dZone;
    time_t tEvent;
    int quad;

    // correct for DST

    if( bDST )
        dZone = nStnZone - 1;
    else
        dZone = nStnZone;

    // convert all measure to radians

    radians( dStnLat );
    radians( dStnLong );
    dZone *= PI/12;       // equiv to 15 degrees/hour

    if( nRiseSet == SUNRISE )
        q = QUAD1;
    else
        q = QUAD3;

    // approximate time
    dApproxTime = g_Time.YearDay() + ( (q + dStnLong)/QUAD4 );

    // solar mean anomaly
    dMeanAnomaly = dApproxTime * 0.017202 - 0.0574039;

    // solar true longitude
    dSolarLong = ( dMeanAnomaly + 0.0334405 * sin( dMeanAnomaly ) ) +
        3.49066e-04 * sin( 2*dMeanAnomaly ) + 4.93289;

    // quadrant determination
    dSolarLong = SunNormalize( dSolarLong );

    t = dSolarLong/QUAD1;

    if( t == (int) t )
        dSolarLong += 4.84814e-6;

    if( dSolarLong > QUAD3 )
        quad = 2;
    else if( dSolarLong <= QUAD1 )
        quad = 0;
    else
        quad = 1;

    // solar right ascension
    dAscension = atan( 0.91746 * tan( dSolarLong ) );

    // quadrant adjust
    switch( quad )
    {
        case 1:
            dAscension += QUAD2;
            break;

        case 2:
            dAscension += QUAD4;
    }

    // solar dDeclinationation
    dDeclination = 0.39782 * sin( dSolarLong );
    dDeclination = atan( dDeclination/sqrt( -dDeclination * dDeclination + 1) );

    // coordinate conversion
    s = (-0.0145439 - ( sin(dDeclination) * sin(dStnLat) )) / ( cos(dDeclination) * cos(dStnLat) );

    // null phenomenon
    if( fabs(s) > 1 )
        return 0;

    // adjustment

    s = QUAD1 - atan(s/sqrt( -s * s + 1));

    if( nRiseSet == SUNRISE )
        s = QUAD4 - s;

    // loval apparent time
    t = s + dAscension - (0.0172028 * dApproxTime) - 1.73364;

    // actual local time
    t = t + dStnLong - dZone;
    t = ( 3.81972 * SunNormalize(t) * MINUTE_SEC );
    tEvent = (time_t) t;
    if( t - (time_t) t )
        tEvent++;

    return tEvent;
}

