/***************************************************************************
 *   Copyright (C) 2005 by John Bennett   *
 *   n4xi@arrl.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

//
// wxcalc.cpp - handles calculations for derived weather data
//

#define _WXCALC_CPP_

#include <math.h>

#include "wxcalc.h"



//////////////////////////////////////////////////////////////////////////////
// CalculateFields - calculates and fills calculated fields in nodedata
// structure.

void CalculateFields( nodedata * pData )
{
    // dew point and heat index
    if( pData->hcr )
    {
        // dew point
        pData->nDewPoint = DewPoint( pData->tcr, pData->hcr );

        // heat index
        if( pData->tcr < 68 )
            pData->nHeatIndex = 0;
        else
            pData->nHeatIndex = HeatIndex( pData->tcr, pData->hcr );

        // fog if rel humidity > 90% and wind velocity is < 5 MPH
        if( (pData->hcr > 90) && (pData->wac.spd < 5) )
            pData->bFog = TRUE;
        else
            pData->bFog = FALSE;
    }
    else
    {
        pData->nDewPoint = 0;
        pData->nHeatIndex = 0;
        pData->bFog = FALSE;
    }

    // Wind Chill Average
    pData->wca = Chill( pData->tcr, pData->wac.spd );
    // Wind Chill Gust
    pData->wcg = Chill( pData->tcr, pData->wgc.spd );
}



///////////////////////////////////////////////////////////////////////////////
// DewPoint - return dewpoint from temperature and relative humidity.
// Temperature is handed off to routine in degrees Fahrenheit.
//

int DewPoint( int nTemp, int nRH )
{
    double dVp;
    double dLogVp;     // vapor pressure
    double dTemp;
    double dDewPtC;
    double dDewPtF;

    dTemp = ( (double)nTemp - 32 ) * 5/9;   // convert to celsius

    dVp = ( 6.11 * pow(10, ((7.5 * dTemp)/(237.7 + dTemp)) ) ) * (double) nRH/100;
    dLogVp = log( dVp );

    dDewPtC = (-430.22 + (237.7*dLogVp))/( 19.08 - dLogVp );

    dDewPtF = (dDewPtC * 9/5 ) + 32;  // convert to Fahrenheit
    if( (dDewPtF - (int) dDewPtF ) < 0.5 )
        return (int) dDewPtF;
    else
        return (int) dDewPtF + 1;
}


///////////////////////////////////////////////////////////////////////////////
// HeatIndex - return heat index from temperature and relative humidity.
// Formula only valid if temperature is 68F or above.
//

int HeatIndex( int nTemp, int nRH )
{
    double dHeatIndex;
    double dTemp2;
    double dTemp3;
    double dRH2;
    double dRH3;

    dTemp2 = (double) nTemp * (double) nTemp;
    dTemp3 = dTemp2 * (double) nTemp;


    dRH2 = (double) nRH * (double) nRH;
    dRH3 = dRH2 * (double) nRH;

    dHeatIndex = 16.923 + 0.185212 * (double) nTemp + 5.37941 * (double) nRH - 0.100354 * (double) nTemp *
        (double) nRH + 0.941695e-2 * dTemp2 + 0.728898e-2 * dRH2 + 0.345372e-3 * dTemp2 * (double) nRH -
        0.81497e-3 * (double) nTemp * dRH2 + 0.102102e-4 * dTemp2 * dRH2 - 0.38646e-4 * dTemp3 + 0.291583e-4 *
        dRH3 + 0.142721e-5 * dTemp3 * (double) nRH + 0.197483e-6 * (double) nTemp * dRH3 - 0.218429e-7 *
        dTemp3 * dRH2 + 0.843296e-9 * dTemp2 * dRH3 - 0.481975e-10 * dTemp3 * dRH3;

    return (int) dHeatIndex;
}


//////////////////////////////////////////////////////////////////////////////
//  Wind Chill (F) = 35.74 + 0.6215 * T - 35.75 * V^0.16 + 0.4275 * T * V^0.16
//  Formula only valid if 50 degrees F or lower.
//

int Chill( int nTemp, int nWind )
{
    double dResult;
    double dWind;

    if( nTemp > 50 )
        return nTemp;

    if( nWind < 3 )
        return nTemp;

    dWind = pow( (double) nWind, 0.16 );
    dResult = 35.74 + 0.6251 * (double) nTemp - 35.75 * dWind + 0.4275 * (double) nTemp * dWind;

    return (int) floor( dResult );
}


///////////////////////////////////////////////////////////////////////////////
// HeatDays - calculate number of heating degree days
//

int HeatDays( int nMeanTemp )
{
    int nDays;

    if( nMeanTemp == 999 )
        return 0;

    if( nMeanTemp > 65 )
        return 0;
    else
        nDays = 65 - nMeanTemp;

    return nDays;
}


///////////////////////////////////////////////////////////////////////////////
// CoolDays - calculate number of cooling degree days
//

int CoolDays( int nMeanTemp )
{
    int nDays;

    if( nMeanTemp == 999 )
        return 0;

    if( nMeanTemp < 65 )
        return 0;
    else
        nDays = nMeanTemp - 65;

    return nDays;
}

