/***************************************************************************
 *   Copyright (C) 2005 by John Bennett   *
 *   n4xi@arrl.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

//
// alarm.cpp - handles alarm situation when specified input paramters exceed
// a given threshold.  While at or above the threshold, it will write critical
// real-time data to the backend database server.
//
//  Data to be written: call_ssid,dtg,wacs,wacd,wgcs,wgcd,pca,pri,prh,bcr,tcr,
//                      ecc,eqc,exc,ehc,rca,rha,rda,rmax,rmaxcnt
//
//  Data points to determine threshold: wac, wgc, ecc, rca
//
//  If a threshold is exceeded on a five minute read, data is written to the
//  database and notification is sent (if so configured).
//

#define _ALARM_CPP_

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include <ctype.h>
#include <unistd.h>
#include <libpq-fe.h>

// libwxn
#include <logmsg.h>
#include <tz.h>
#include <cconfig.h>
#include <strxcpy.h>
#include <makelc.h>

#include "hconfig.h"
#include "hglobal.h"
#include "wxdata.h"
#include "sensors.h"
#include "alarm.h"
#include "email.h"


#define MSG_SIZE 1024


// local prototype
void WriteAlarm( void );
int AddMessage( char * pBuf, const char * pMsg );
int MakeTextMessage( char * psMsg, int nType, char * psSubject );
int MakePagerMessage( char * psMsg, int nType, char * psSubject );

// local variables

static char g_szThisAlarmUTC[32];

static int g_nGustSpeed = 0;
static int g_nGustDir = 0;
static int g_nAvgSpeed = 0;
static int g_nAvgDir = 0;
static int g_nInstPrecip = 0;
static int g_nPrecipHour = 0;
static int g_n24HrPrecip = 0;
static int g_nLoBarom = 0;
static unsigned g_uEcc = 0;
static int g_nHiRad = 0;
static int g_nHiTemp = 0;
static int g_nLoTemp = 0;

static BOOL g_b24HrPrecip = FALSE;
static BOOL g_bPrecipHour = FALSE;

// thread related variables/functions
void * Notify( void * pParam );
void CheckAlarmThread( void );
static pthread_t g_hAlarmThread;
static BOOL g_bAlarmThread = FALSE;


//////////////////////////////////////////////////////////////////////////////
// DataAlarm - check for data alarm conditions and act accordingly
//

void DataAlarm( void )
{

    int nType;
    int nStatus;

    nType = 0;

    ////////////// Wind Gust
    if( g_Local.Data.wgc.spd > g_nGustSpeed )
    {
        if( g_Local.Data.wgc.spd >= g_HostConfig.Alarm.nHiWindGust )
        {
            g_nGustSpeed = g_Local.Data.wgc.spd;
            g_nGustDir = g_Local.Data.wgc.dir;
            nType |= WIND_ALARM;
        }
    }

    ////////////// Wind Average
    if( g_Local.Data.wac.spd > g_nAvgSpeed )
    {
        if( g_Local.Data.wac.spd >= g_HostConfig.Alarm.nHiWindGust )
        {
            g_nAvgSpeed = g_Local.Data.wac.spd;
            g_nAvgDir = g_Local.Data.wac.dir;
            nType |= WIND_ALARM;
        }
    }

    ////////////// Precip Rate - hour
    if( g_Local.Data.prh > g_nPrecipHour )
    {
        if( !g_bPrecipHour )
        {
            g_bPrecipHour = TRUE;
            if( g_Local.Data.prh >= g_HostConfig.Alarm.nHourPrecip )
            {
                g_nPrecipHour = g_Local.Data.prh;
                nType |= PRECIP_HR_ALARM;
            }
        }
    }

    ////////////// Precip - 24 hour
    if( g_Local.Data.p24 > g_n24HrPrecip )
    {
        if( !g_b24HrPrecip )
        {
            g_b24HrPrecip = TRUE;
            if( g_Local.Data.p24 >= g_HostConfig.Alarm.n24HrPrecip )
            {
                g_n24HrPrecip = g_Local.Data.p24;
                nType |= PRECIP_24_ALARM;
            }
        }
    }

    ////////////// Low Barometric Pressure
    if( g_Sensors.bOwBarom || g_Sensors.bBarom )
    {
        if( g_Local.Data.bcr < g_nLoBarom )
        {
            if( g_Local.Data.bcr <= g_HostConfig.Alarm.nLoBarom )
            {
                g_nLoBarom = g_Local.Data.bcr;
                nType |= BAROM_ALARM;
            }
        }
    }

    ////////////// EMP
    if( g_Sensors.bOwEmp )
    {
        if( g_Local.Data.ecc > g_uEcc )
        {
            if( g_Local.Data.ecc >= g_HostConfig.Alarm.nHiEmp )
            {
                g_uEcc = g_Local.Data.ecc;
                nType |= EMP_ALARM;
            }
        }
    }

    ////////////// Background Radiation
    if( g_Sensors.bOwRad )
    {
        if( g_Local.Data.rca > g_nHiRad )
        {
            if( g_Local.Data.rca >= g_HostConfig.Alarm.nHiRad )
            {
                g_nHiRad = g_Local.Data.rca;
                nType |= RAD_ALARM;
            }
        }
    }

    ////////////// High Temperature
    if( g_Local.Data.tcr > g_nHiTemp )
    {
        if( g_Local.Data.tcr >= g_HostConfig.Alarm.nHiTemp )
        {
            g_nHiTemp = g_Local.Data.tcr;
            nType |= HTEMP_ALARM;
        }
    }

    ////////////// Low Temperature
    if( g_Local.Data.tcr < g_nLoTemp )
    {
        if( g_Local.Data.tcr <= g_HostConfig.Alarm.nLoTemp )
        {
            g_nLoTemp = g_Local.Data.tcr;
            nType |= LTEMP_ALARM;
        }
    }

    //////////////////////////////////

    if( nType )
    {
        CheckAlarmThread();
        WriteAlarm();
        if( g_HostConfig.email.bEnable )
        {
            nStatus = pthread_create( &g_hAlarmThread, NULL, &Notify, &nType );
            if( nStatus )
                g_log.PostMsg( MSG_THREAD_START_FAIL, "alarm" );
        }
    }
}


//////////////////////////////////////////////////////////////////////////////
// ResetAlarms - called at the top of each hour
//

void ResetAlarms( void )
{

    g_nGustSpeed = 0;
    g_nGustDir = 0;

    g_nAvgSpeed = 0;
    g_nAvgDir = 0;

    g_nInstPrecip = 0;
    g_nPrecipHour = 0;
    g_n24HrPrecip = g_Local.Data.p24;

    g_nLoBarom = g_Local.Data.bcr;
    g_uEcc = 0;
    g_nHiRad = 0;
    g_nHiTemp = g_Local.Data.tcr;
    g_nLoTemp = g_Local.Data.tcr;

    g_b24HrPrecip = FALSE;
    g_bPrecipHour = FALSE;

}


//////////////////////////////////////////////////////////////////////////////
// CheckAlarmThread - if thread running, wait for termination and return.
//

void CheckAlarmThread( void )
{
    if( g_bAlarmThread )
    {
        pthread_join( g_hAlarmThread, NULL );
        sleep( 5 );     // allow everything to synchronize
    }
}


//////////////////////////////////////////////////////////////////////////////
// WriteAlarm - write alarm data to table
//

void WriteAlarm( void )
{
    PGconn * pConn;
    PGresult * pResult;
    char szExec[384];
    char szDtgLocal[32];
    char szDtgZulu[32];

    *g_szThisAlarmUTC = 0;

    pConn = PQconnectdb( "dbname=wxn" );
    if( pConn == NULL )
        return;
    if( PQstatus( pConn ) != CONNECTION_OK )
    {
        PQfinish( pConn );
        return;
    }

    // create the record
    g_Time.CurrentLocalTimestamp( szDtgLocal );
    g_Time.CurrentZuluTimestamp( szDtgZulu );
    strcpy( g_szThisAlarmUTC, szDtgZulu );
    sprintf( szExec, "INSERT INTO alarm ( call_ssid, dtg_local, dtg_zulu ) VALUES( '%s', '%s', '%s' )",
        g_Local.Info.szCallSsid, szDtgLocal, szDtgZulu );
    pResult = PQexec( pConn, szExec );
    if( PQresultStatus( pResult ) != PGRES_COMMAND_OK )
    {
        PQclear( pResult );
        PQfinish( pConn );
        return;
    }
    PQclear( pResult );

    sprintf( szExec, "BEGIN work" );
    pResult = PQexec( pConn, szExec );
    PQclear( pResult );

    // wind
    sprintf( szExec, "UPDATE alarm SET wgc_spd=%i, wgc_dir=%i, wac_spd=%i, wac_dir=%i "
        "WHERE call_ssid='%s' AND dtg_zulu='%s'",
        g_Local.Data.wgc.spd, g_Local.Data.wgc.dir, g_Local.Data.wac.spd, g_Local.Data.wac.dir,
        g_Local.Info.szCallSsid, szDtgZulu );
    pResult = PQexec( pConn, szExec );
    PQclear( pResult );

    // precip, temp
    sprintf( szExec, "UPDATE alarm SET pca=%i, pri=%i, prh=%i, p24=%i, tcr=%i "
        "WHERE call_ssid='%s' AND dtg_zulu='%s'",
        g_Local.Data.pca, g_Local.Data.pri, g_Local.Data.prh, g_Local.Data.p24, g_Local.Data.tcr,
        g_Local.Info.szCallSsid, szDtgZulu );
    pResult = PQexec( pConn, szExec );
    PQclear( pResult );

    // barometer
    if( g_Sensors.bBarom || g_Sensors.bOwBarom )
    {
        sprintf( szExec, "UPDATE alarm SET bcr=%i WHERE call_ssid='%s' AND dtg_zulu='%s'",
            g_Local.Data.bcr, g_Local.Info.szCallSsid, szDtgZulu );
        pResult = PQexec( pConn, szExec );
        PQclear( pResult );
    }

    // humidity
    if( g_Sensors.bOutsideHumid || g_Sensors.bOwOutsideHumid )
    {
        sprintf( szExec, "UPDATE alarm SET hcr=%i WHERE call_ssid='%s' AND dtg_zulu='%s'",
            g_Local.Data.hcr, g_Local.Info.szCallSsid, szDtgZulu );
        pResult = PQexec( pConn, szExec );
        PQclear( pResult );
    }

    // emp
    if( g_Sensors.bOwEmp )
    {
        sprintf( szExec, "UPDATE alarm SET ecc=%i, eqc=%i, exc=%i, ehc=%i, edc=%i "
            "WHERE call_ssid='%s' AND dtg_zulu='%s'",
            g_Local.Data.ecc, g_Local.Data.eqc, g_Local.Data.exc, g_Local.Data.ehc, g_Local.Data.edc,
            g_Local.Info.szCallSsid, szDtgZulu );
        pResult = PQexec( pConn, szExec );
        PQclear( pResult );
    }

    // radiation
    if( g_Sensors.bOwRad )
    {
        sprintf( szExec, "UPDATE alarm SET rca=%i, rha=%i, rda=%i, rmax=%i, ralarm_count=%i, ralarm_trip=%i "
            "WHERE call_ssid='%s' AND dtg_zulu='%s'",
            g_Local.Data.rca, g_Local.Data.rha, g_Local.Data.rda, g_Local.Data.rmax,
            g_Local.Data.ralarm_count, g_HostConfig.Alarm.nHiRad,
            g_Local.Info.szCallSsid, szDtgZulu );
        pResult = PQexec( pConn, szExec );
        PQclear( pResult );
    }

    sprintf( szExec, "COMMIT work" );
    pResult = PQexec( pConn, szExec );
    PQclear( pResult );

    PQfinish( pConn );
}


//////////////////////////////////////////////////////////////////////////////
// Notify - email recipients in email table alarm data
//

void * Notify( void * pParam )
{
    PGconn * pConn;
    PGresult * pResult;
    char szExec[384];
    char szOnly[16];
    char szTxtSubject[64], szPagerSubject[64];
    char szTxtMsg[MSG_SIZE+1], szPagerMsg[128];
    mailform TxtMail, PagerMail;
    int nRows;
    int nTxtLen, nPagerLen;
    int nType;
    char szMode[2];
    BOOL bSendThis;
    char * pS;
    int i;

    g_bAlarmThread = TRUE;

    nType = *(int *)pParam;

    nTxtLen = MakeTextMessage( szTxtMsg, nType, szTxtSubject );
    nPagerLen = MakePagerMessage( szPagerMsg, nType, szPagerSubject );

    ////////////// common email header info, message body

    if( *g_HostConfig.email.sFromName )
    {
        strcpy( TxtMail.sFromName, g_HostConfig.email.sFromName );
        strcpy( PagerMail.sFromName, g_HostConfig.email.sFromName );
    }
    else if( *g_ComConfig.szAlias )
    {
        strcpy( TxtMail.sFromName, g_ComConfig.szAlias );
        strcpy( PagerMail.sFromName, g_ComConfig.szAlias );
    }
    else
    {
        strcpy( TxtMail.sFromName, "wxn" );
        strcpy( PagerMail.sFromName, "wxn" );
    }

    strcpy( TxtMail.sFromAcct, g_HostConfig.email.sFromAcct );
    strcpy( TxtMail.sFromHost, g_HostConfig.email.sHost );
    strcpy( TxtMail.sReplyTo, g_HostConfig.email.sReplyTo );
    sprintf( TxtMail.sSubject, "wx alarm: %s", szTxtSubject );
    TxtMail.pBody = szTxtMsg;
    TxtMail.nBodyLen = nTxtLen;

    strcpy( PagerMail.sFromAcct, g_HostConfig.email.sFromAcct );
    strcpy( PagerMail.sFromHost, g_HostConfig.email.sHost );
    strcpy( PagerMail.sReplyTo, g_HostConfig.email.sReplyTo );
    sprintf( PagerMail.sSubject, "wx alarm: %s", szPagerSubject );
    PagerMail.pBody = szPagerMsg;
    PagerMail.nBodyLen = nPagerLen;

    ////////////// email recipients

    pConn = PQconnectdb( "dbname=wxn" );
    if( pConn == NULL )
    {
        g_bAlarmThread = FALSE;
        return NULL;
    }
    if( PQstatus( pConn ) != CONNECTION_OK )
    {
        PQfinish( pConn );
        g_bAlarmThread = FALSE;
        return NULL;
    }

    sprintf( szExec, "SELECT recip,host,name,mtype,alerts FROM email" );
    pResult = PQexec( pConn, szExec );
    if( PQresultStatus( pResult ) == PGRES_TUPLES_OK )
    {
        nRows = PQntuples( pResult ) ;
        if( nRows )
        {
            for( i=0; i<nRows; i++ )
            {
                strxcpy( szMode, PQgetvalue( pResult, i, 3 ), 1 );
                *szMode = tolower( *szMode );
                if( !PQgetisnull( pResult, i, 4 ) )
                {
                    bSendThis = FALSE;
                    strcpy( szOnly, PQgetvalue( pResult, i, 4 ) );
                    makelc( szOnly );
                    pS = szOnly;
                    while( *pS )
                    {
                        bSendThis = TRUE;
                        if( *pS == '*')
                            break;
                        else if( *pS == 't' && (nType & HTEMP_ALARM || nType & LTEMP_ALARM) )
                            break;
                        else if( *pS == 'w' && nType & WIND_ALARM )
                            break;
                        else if( *pS == 'h' && nType & PRECIP_HR_ALARM )
                            break;
                        else if( *pS == 'd' && nType & PRECIP_24_ALARM )
                            break;
                        else if( *pS == 'b' && nType & BAROM_ALARM )
                            break;
                        else if( *pS == 'e' && nType & EMP_ALARM )
                            break;
                        else if( *pS == 'r' && nType & RAD_ALARM )
                            break;
                        else
                            bSendThis = FALSE;
                        pS++;
                    }
                    if( !bSendThis )
                        continue;   // this recipient does not want anything alarmed here
                }
                if( *szMode == 't' )
                {
                    strcpy( TxtMail.sToAcct, PQgetvalue( pResult, i, 0 ) );
                    strcpy( TxtMail.sToHost, PQgetvalue( pResult, i, 1 ) );
                    if( PQgetisnull( pResult, i, 2 ) )
                        *TxtMail.sToName = 0;
                    else
                        strcpy( TxtMail.sToName, PQgetvalue( pResult, i, 2 ) );
                    if ( !email( TxtMail ) )
                    {
                        sprintf( szExec, "%s <%s@%s>", TxtMail.sToName, TxtMail.sToAcct, TxtMail.sToHost );
                        g_log.PostMsg( HOSTMSG_EMAIL_ALERT_FAIL, szExec );
                    }
                }
                else
                {
                    strcpy( PagerMail.sToAcct, PQgetvalue( pResult, i, 0 ) );
                    strcpy( PagerMail.sToHost, PQgetvalue( pResult, i, 1 ) );
                    if( PQgetisnull( pResult, i, 2 ) )
                        *TxtMail.sToName = 0;
                    else
                        strcpy( PagerMail.sToName, PQgetvalue( pResult, i, 2 ) );
                    if ( !email( PagerMail ) )
                    {
                        sprintf( szExec, "%s <%s@%s>", PagerMail.sToName, PagerMail.sToAcct, PagerMail.sToHost );
                        g_log.PostMsg( HOSTMSG_EMAIL_ALERT_FAIL, szExec );
                    }
                }
            }
        }
    }

    PQclear( pResult );
    PQfinish( pConn );

    g_bAlarmThread = FALSE;
    return NULL;
}

//////////////////////////////////////////////////////////////////////////////
// MakeTextMessage
//

int MakeTextMessage( char * psMsg, int nType, char * psSubject )
{
    char szString[128];
    int nLen;

    nLen = 0;
    *psMsg = 0;
    *psSubject = 0;
    nLen += AddMessage( psMsg, "Wxn Email Server Alert\r\n" );
    nLen += AddMessage( psMsg, "DO NOT REPLY (automated message). Click on 'Reply-To' if you need to contact the sysop.\r\n" );
    nLen += AddMessage( psMsg, "The following parameter(s) tripped this alert: " );
    if( nType & PRECIP_HR_ALARM || nType & PRECIP_24_ALARM )
        strcat( psSubject, "precip " );
    if( nType & EMP_ALARM )
        strcat( psSubject, "emp " );
    if( nType & WIND_ALARM )
        strcat( psSubject, "wind " );
    if( nType & BAROM_ALARM )
        strcat( psSubject, "barom " );
    if( nType & HTEMP_ALARM || nType & LTEMP_ALARM )
        strcat( psSubject, "temp " );
    if( nType & RAD_ALARM )
        strcat( psSubject, "rad" );
    nLen += AddMessage( psMsg, psSubject );
    nLen += AddMessage( psMsg, "\r\n" );

    //// location
    sprintf( szString, "Location: %s, %s (%s)\r\n",
        g_Local.Info.szCity, g_Local.Info.szState, g_Local.Info.szCounty );
    nLen += AddMessage( psMsg, szString );
    sprintf( szString, "Coordinates: %.2f N, %.2f W\r\n",
        g_Local.Info.dLatitude, g_Local.Info.dLongitude );
    nLen += AddMessage( psMsg, szString );

    //// dtg_zulu
    sprintf( szString, "Date/Time of alert: %s UTC\r\n", g_szThisAlarmUTC );
    nLen += AddMessage( psMsg, szString );

    //// temp
    sprintf( szString, "Temperature = %i\r\n", g_Local.Data.tcr );
    nLen += AddMessage( psMsg, szString );

    //// precip
    // pca
    sprintf( szString, "PRECIP:\tTotal since midnight local time = %.2f\"\r\n", g_Local.Data.pca/100.00 );
    nLen += AddMessage( psMsg, szString );
    // p24
    sprintf( szString, "\tTotal last 24 hours = %.2f\"\r\n", g_Local.Data.p24/100.00 );
    nLen += AddMessage( psMsg, szString );
    // prh
    sprintf( szString, "\tHourly Rate = %.2f\"\r\n", g_Local.Data.prh/100.00 );
    nLen += AddMessage( psMsg, szString );

    //// wind
    // wgc_spd, wgc_dir
    sprintf( szString, "WIND:\tGust Dir/Speed: %i deg @ %i mph\r\n", g_Local.Data.wgc.dir, g_Local.Data.wgc.spd );
    nLen += AddMessage( psMsg, szString );
    // wac_spd, wac_dir
    sprintf( szString, "\tAvg Dir/Speed: %i deg @ %i mph\r\n", g_Local.Data.wac.dir, g_Local.Data.wac.spd );
    nLen += AddMessage( psMsg, szString );

    // barometer
    if( g_Sensors.bBarom || g_Sensors.bOwBarom )
    {
        sprintf( szString, "Barometric Pressure: %0.2f\" (%.1f mb)\r\n", g_Local.Data.bcr/333.83, g_Local.Data.bcr/10.0 );
        nLen += AddMessage( psMsg, szString );
    }
    else
        nLen += AddMessage( psMsg, "NO barometric pressure sensor\r\n" );

    // humidity
    if( g_Sensors.bOutsideHumid || g_Sensors.bOwOutsideHumid )
    {
        sprintf( szString, "Humidity = %i%%\r\n", g_Local.Data.hcr );
        nLen += AddMessage( psMsg, szString );
    }
    else
        nLen += AddMessage( psMsg, "NO humidity sensor\r\n" );

    // emp
    if( g_Sensors.bOwEmp )
    {
        sprintf( szString, "EMP 5/15/30/60/Day = %i / %i / %i / %i / %i\r\n",
            g_Local.Data.ecc,    // ecc
            g_Local.Data.eqc,    // eqc
            g_Local.Data.exc,    // exc
            g_Local.Data.ehc,    // ehc
            g_Local.Data.edc     // edc
            );
        nLen += AddMessage( psMsg, szString );
    }
    else
        nLen += AddMessage( psMsg, "NO emp sensor\r\n" );

    // radiation
    if( g_Sensors.bOwRad )
    {
        nLen += AddMessage( psMsg, "Background Radiation Monitor\r\n" );
        sprintf( szString, "\tCurrent/Hourly/Daily Avgs = %i / %i / %i (uRtgns)\r\n",
            g_Local.Data.rca,    // rca
            g_Local.Data.rha,    // rha
            g_Local.Data.rda     // rda
            );
        nLen += AddMessage( psMsg, szString );

        sprintf( szString, "\tMaximum for the day: %i uRtgns\r\n", g_Local.Data.rmax );    // rmax
        nLen += AddMessage( psMsg, szString );

        sprintf( szString, "\tAlarm trips for day / Alarm trip-point: %i / %i (uRtgns)\r\n",
            g_Local.Data.ralarm_count,    // ralarm_count
            g_HostConfig.Alarm.nHiRad      // ralarm_trip
            );
        nLen += AddMessage( psMsg, szString );
    }
    else
        nLen += AddMessage( psMsg, "NO background radiation sensor\r\n" );

    nLen += AddMessage( psMsg, "*** End Of Message\r\n" );

    // message creation complete

    return nLen;
}


//////////////////////////////////////////////////////////////////////////////
// MakePagerMessage
//

int MakePagerMessage( char * psMsg, int nType, char * psSubject )
{
    char szString[128];
    int nLen;

    nLen = 0;
    *psMsg = 0;
    *psSubject = 0;

    //// location/dtg
    sprintf( szString, "@%s,%2.2i%2.2i%2.2i%2.2i\r\n",
        g_Local.Info.szFips,
        g_Time.Month(), g_Time.Day(),
        g_Time.Hour(), g_Time.Minute() );
    nLen += AddMessage( psMsg, szString );

    //// temp
    if( nType & HTEMP_ALARM || nType & LTEMP_ALARM )
    {
        strcat( psSubject, "temp " );
        sprintf( szString, "tcr=%i\r\n", g_Local.Data.tcr );
        nLen += AddMessage( psMsg, szString );
    }

    //// precip
    if( nType & PRECIP_HR_ALARM || nType & PRECIP_24_ALARM )
    {
        strcat( psSubject, "precip " );
        sprintf( szString, "pca=%i,prh=%i,p24=%i\r\n",
            g_Local.Data.pca, g_Local.Data.prh, g_Local.Data.p24 );
        nLen += AddMessage( psMsg, szString );
    }

    //// wind
    if( nType & WIND_ALARM )
    {
        strcat( psSubject, "wind " );
        sprintf( szString, "wgc=%i@%i,wac=%i@%i\r\n",
            g_Local.Data.wgc.dir, g_Local.Data.wgc.spd, g_Local.Data.wac.dir, g_Local.Data.wac.spd );
        nLen += AddMessage( psMsg, szString );
    }

    // barometer
    if( g_Sensors.bBarom || g_Sensors.bOwBarom )
    {
        if( nType & BAROM_ALARM )
        {
            strcat( psSubject, "barom " );
            sprintf( szString, "bcr=%i\r\n", g_Local.Data.bcr/10 );
            nLen += AddMessage( psMsg, szString );
        }
    }

    // emp
    if( g_Sensors.bOwEmp )
    {
        if( nType & EMP_ALARM )
        {
            strcat( psSubject, "emp " );
            sprintf( szString, "ecc=%u,eqc=%u,ehc=%u\r\n",
                g_Local.Data.ecc, g_Local.Data.eqc, g_Local.Data.ehc );
            nLen += AddMessage( psMsg, szString );
        }
    }

    // radiation
    if( g_Sensors.bOwRad )
    {
        if( nType & RAD_ALARM )
        {
            strcat( psSubject, "rad" );
            sprintf( szString, "rca=%i,rha=%i,rmax=%i\r\n",
                g_Local.Data.rca, g_Local.Data.rha, g_Local.Data.rmax );
            nLen += AddMessage( psMsg, szString );
        }
    }

    return nLen;
}


//////////////////////////////////////////////////////////////////////////////
// ReadAlarmConfig
//

void ReadAlarmConfig( void )
{
    PGconn * pConn;
    PGresult * pResult;
    char szExec[384];

    pConn = PQconnectdb( "dbname=wxn" );
    if( pConn == NULL )
        return;
    if( PQstatus( pConn ) != CONNECTION_OK )
    {
        PQfinish( pConn );
        return;
    }

    sprintf( szExec, "SELECT prh,p24,wgh,wah,thr,tlr,blr,ecc,rad FROM alarmpoints WHERE call_ssid='%s'",
        g_Local.Info.szCallSsid );
    pResult = PQexec( pConn, szExec );
    if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
    {
        g_HostConfig.Alarm.nHourPrecip = atoi( PQgetvalue( pResult, 0, 0 ) );
        g_HostConfig.Alarm.n24HrPrecip = atoi( PQgetvalue( pResult, 0, 1 ) );
        g_HostConfig.Alarm.nHiWindGust = atoi( PQgetvalue( pResult, 0, 2 ) );
        g_HostConfig.Alarm.nHiWindAvg = atoi( PQgetvalue( pResult, 0, 3 ) );
        g_HostConfig.Alarm.nHiTemp = atoi( PQgetvalue( pResult, 0, 4 ) );
        g_HostConfig.Alarm.nLoTemp = atoi( PQgetvalue( pResult, 0, 5 ) );
        g_HostConfig.Alarm.nLoBarom = atoi( PQgetvalue( pResult, 0, 6 ) ) * 10; // stored in 0.1 mb
        g_HostConfig.Alarm.nHiEmp = atoi( PQgetvalue( pResult, 0, 7 ) );
        g_HostConfig.Alarm.nHiRad = atoi( PQgetvalue( pResult, 0, 8 ) );
    }
    else    // record does not exist
    {
        g_HostConfig.Alarm.nInstPrecip = PRI_DEFAULT;
        g_HostConfig.Alarm.nHourPrecip = PRH_DEFAULT;
        g_HostConfig.Alarm.n24HrPrecip = P24_DEFAULT;
        g_HostConfig.Alarm.nHiWindGust = WGH_DEFAULT;
        g_HostConfig.Alarm.nHiWindAvg = WAH_DEFAULT;
        g_HostConfig.Alarm.nHiTemp = THR_DEFAULT;
        g_HostConfig.Alarm.nLoTemp = TLR_DEFAULT;
        g_HostConfig.Alarm.nLoBarom = BLR_DEFAULT * 10; // stored in 0.1 mb
        g_HostConfig.Alarm.nHiEmp = EMP_DEFAULT;
        g_HostConfig.Alarm.nHiRad = RAD_DEFAULT;

        PQclear( pResult );
        sprintf( szExec, "INSERT INTO alarmpoints( call_ssid ) VALUES( '%s' )", g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );
        if( PQresultStatus( pResult ) != PGRES_COMMAND_OK )
        {
            PQclear( pResult );
            PQfinish( pConn );
            return;
        }
        PQclear( pResult );

        // set defaults
        sprintf( szExec, "UPDATE alarmpoints SET pri=%i,prh=%i,p24=%i,wgh=%i,wah=%i,thr=%i,tlr=%i,blr=%i,ecc=%i,rad=%i "
            "WHERE call_ssid='%s'",
            PRI_DEFAULT,PRH_DEFAULT, P24_DEFAULT, WGH_DEFAULT, WAH_DEFAULT, THR_DEFAULT,
            TLR_DEFAULT, BLR_DEFAULT, EMP_DEFAULT, RAD_DEFAULT,
            g_Local.Info.szCallSsid );
        pResult = PQexec( pConn, szExec );

    }

    PQclear( pResult );
    PQfinish( pConn );
}


//////////////////////////////////////////////////////////////////////////////
// AddMessage
//

int AddMessage( char * pBuf, const char * pMsg )
{
    int nLen;
    char * pS;
    char * pD;

    nLen = 0;

    pS = (char *) pMsg;
    pD = pBuf;

    while( *pD )
        pD++;

    while( *pS )
    {
        *pD++ = *pS++;
        nLen++;
    }

    *pD = 0;

    return nLen;
}

