/***************************************************************************
 *   Copyright (C) 2005 by John Bennett   *
 *   n4xi@arrl.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

//
// davp.cpp - driver for Davis Instruments Vantage Pro weather station
//

#define _DAVP_CPP_

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <pthread.h>
#include <unistd.h>

// libwxn
#include <logmsg.h>

#include "precip.h"
#include "barom.h"
#include "humid.h"
#include "temp.h"
#include "wind.h"
#include "wxport.h"
#include "sensors.h"
#include "rates.h"
#include "hglobal.h"
#include "crcdav.h"
#include "fixvalue.h"
#include "wxdata.h"
#include "davp.h"


#ifndef byte_t
    typedef unsigned char byte_t;
#endif

#ifndef word_t
    typedef unsigned short word_t;
#endif

#define BYTE_ERROR  0xff
#define WORD_ERROR  0x7fff  // signed word error
#define UWORD_ERROR 0xffff  // unsigned word error

#define ACK         0x06
#define NACK        '!'
#define DAV_ERROR   0x80

#define ACK_SIZE        1
#define BLOCK_SIZE      99

// Offsets within the block

#define TIN     9   // 2 bytes (9, 10)
#define TOUT    12  // 2 bytes (12, 13)
#define WSPD    14  // 1 byte
#define WDIR    16  // 2 bytes (16, 17)
#define BAR     7   // 2 bytes (7, 8)
#define HIN     11  // 1 byte
#define HOUT    33  // 1 byte
#define PDAY    50
#define PMONTH  52
#define PYEAR   54
#define CRC     97


//////////////////////////////////////
// local function prototypes
//////////////////////////////////////

static BOOL GetAck( void );
static BOOL ResetLink( void );
static BOOL GetBlock( byte_t * pBlock );
static BOOL StartLoop( void );

// thread related functions

static BOOL DavPStartThread( BOOL bVerbose );
static void * DavPThread( void * pParam );


//////////////////////////////////////
// global variables @ file scope

// temporary values
static int g_nSpeedAccum = 0;
static int g_nDirAccum = 0;
static int g_nSpeedCount = 0;
static int g_nDirCount = 0;
static int g_nDaVPQ1 = 0;
static int g_nDaVPQ4 = 0;

static int g_nGustSpeed = 0;
static int g_nGustDir = 0;

static int g_nHiTemp;
static int g_nLoTemp;

static byte_t g_DataBlock[BLOCK_SIZE];
static int g_nBlockCount = 0;

static int g_nErrTempOut = 0;
static int g_nErrTempIn = 0;
static int g_nErrWindSpd = 0;
static int g_nErrWindDir = 0;
static int g_nErrHumidityOut = 0;
static int g_nErrHumidityIn = 0;
static int g_nErrBarometer = 0;
static int g_nErrPrecip = 0;
static int g_nErrCrc = 0;


//////////////////////////////////////

pthread_mutex_t mx_DavPLock = PTHREAD_MUTEX_INITIALIZER;

// worker thread
static pthread_t g_hThread;
static BOOL g_bStopThread = FALSE;

////////////////////////////////////////////////
// Davis commands
//

static char g_szLoopDataCmd[] = "LOOP 1\n";
#define LOOPDATA_CMD_LEN sizeof( g_szLoopDataCmd ) - 1

static char g_szMultLoopDataCmd[] = "LOOP 1000\n";
#define MULTLOOPDATA_CMD_LEN sizeof( g_szMultLoopDataCmd ) - 1


//////////////////////////////////////////////////////////////////////////////
// DavPProcessData - read weather station, set hi/low values, etc. Returns
// TRUE success, FALSE on failure.
//

BOOL DavPProcessData( void )
{
    int nVal;
    cwinds Wind;

    if( g_nBlockCount == 0 )
    {
         g_log.PostMsg( HOSTMSG_DAVP_BLOCKS, NULL );
         return FALSE;
    }

    pthread_mutex_lock( &mx_DavPLock );

    if( !ResetLink() )
    {
        g_log.PostMsg( HOSTMSG_RESETLINK_FAIL, NULL );
        pthread_mutex_unlock( &mx_DavPLock );
        return FALSE;
    }

    if( !CrcDav( g_DataBlock, BLOCK_SIZE) )
    {
        g_nErrCrc++;
        pthread_mutex_unlock( &mx_DavPLock );
        return FALSE;
    }

    ////////////////////////////////// TEMPERATURE

    nVal = (short int) (g_DataBlock[TOUT] | (g_DataBlock[TOUT+1] << 8));
    if( nVal != WORD_ERROR )
    {

        nVal = FixValue( nVal );
        TempSet( nVal, g_nHiTemp, g_nLoTemp );

        // reset hi/lo temps
        g_nHiTemp = nVal;
        g_nLoTemp = nVal;
    }
    else    // log sensor error
    {
        g_nErrTempOut++;
    }

    // inside
    nVal = (short int) (g_DataBlock[TIN] | (g_DataBlock[TIN+1] << 8));
    if( nVal != WORD_ERROR )
    {
        nVal = FixValue( nVal );
        InsideTempSet( nVal );
    }
    else    // log sensor error
    {
        g_nErrTempIn++;
    }

    ////////////////////////////////// Barometric Pressure

    nVal = (short int) (g_DataBlock[BAR] | (g_DataBlock[BAR+1] << 8));
    if( nVal != UWORD_ERROR )
    {
        nVal = FixValue( nVal );
        nVal = MillibarsT( nVal );
        BaromSet( nVal );
    }
    else    // log sensor error
    {
        g_nErrBarometer++;
    }

    ////////////////////////////////// Humidity

    // outside
    nVal = (short int) g_DataBlock[HOUT];
    if( nVal != BYTE_ERROR )
    {
        if( nVal <= 100 )
            HumidSet( nVal );
    }
    else    // log sensor error
    {
        g_nErrHumidityOut++;
    }

    // inside
    nVal = (short int) g_DataBlock[HIN];
    if( nVal != BYTE_ERROR )
    {
        if( nVal <= 100 )
            InsideHumidSet( nVal );
    }
    else    // log sensor error
    {
        g_nErrHumidityIn++;
    }


    ////////////////////////////////// WIND

    if( g_nSpeedCount )
    {
        //// averages
        Wind.nSpeed = g_nSpeedAccum / g_nSpeedCount;

        if( g_nDaVPQ1 && g_nDaVPQ4 )
        {
            g_nDirAccum += g_nDaVPQ1 * 360;
        }
        Wind.nDir = g_nDirAccum / g_nDirCount;
        if( Wind.nDir > 359 )
            Wind.nDir -= 360;

        //// wind gusts
        if( g_nSpeedCount == 1 )
        {
            Wind.nGustSpeed = Wind.nSpeed;
            Wind.nGustDir = Wind.nDir;
        }
        else    // > 1
        {
            Wind.nGustSpeed = g_nGustSpeed;
            Wind.nGustDir = g_nGustDir;
        }
    }
    else
        memset( &Wind, 0, sizeof( cwinds ) );

    // reset gust speed/dir/accum
    g_nGustSpeed = 0;
    g_nGustDir = 0;
    g_nSpeedAccum = 0;

    g_nDirAccum = 0;
    g_nSpeedCount = 0;
    g_nDirCount = 0;
    g_nDaVPQ1 = 0;
    g_nDaVPQ4 = 0;

    WindSet( &Wind );

    ////////////////////////////////// Precip

    nVal = (short int) (g_DataBlock[PYEAR] | (g_DataBlock[PYEAR+1] << 8) );
    if( nVal != UWORD_ERROR )
    {
        PrecipSet( nVal );
    }
    else    // log sensor error
    {
        g_nErrPrecip++;
    }

    ////////////////////////////////// Clean-Up

    g_nBlockCount = 0;

    StartLoop();

    pthread_mutex_unlock( &mx_DavPLock );

    return TRUE;
}


//////////////////////////////////////////////////////////////////////////////
// GetBlock - retrieves block of data (packet) from Davis wx stn. Returns
// TRUE success, FALSE failure.
//

BOOL GetBlock( byte_t * pBlock )

{
    if( !ResetLink() )
        return FALSE;
    if( !WriteWxStn( g_szLoopDataCmd, LOOPDATA_CMD_LEN ) )
        return FALSE;
    if( !GetAck() )
        return FALSE;
    if( !ReadWxStnBytes( pBlock, BLOCK_SIZE ) )
    {
        g_log.PostMsg( HOSTMSG_READWXSTNBYTES_FAIL, " GetBlock()" );
        return FALSE;
    }

    if( CrcDav(pBlock, BLOCK_SIZE) )
        return TRUE;
    else
        return FALSE;
}


//////////////////////////////////////////////////////////////////////////////
// StartLoop - start Vantage Pro in continuous loop mode until next read.
// Returns TRUE success, FALSE failure.
//

BOOL StartLoop( void )

{
    if( !ResetLink() )
        return FALSE;
    if( !WriteWxStn( g_szMultLoopDataCmd, MULTLOOPDATA_CMD_LEN ) )
        return FALSE;
    if( !GetAck() )
        return FALSE;

    return TRUE;
}


//////////////////////////////////////////////////////////////////////////////
// DavPInit - check connection, determine attached sensors.
//

BOOL DavPInit( BOOL bVerbose )
{
    byte_t DataBlock[BLOCK_SIZE];
    int nVal;
    int i;
    timespec wait;

    wait.tv_sec = 0;
    wait.tv_nsec = 5000;

    //////////////////////// set sensor flags since Vantage Pro has all of these

    g_Sensors.bInsideHumid = TRUE;
    g_Sensors.bOutsideHumid = TRUE;
    g_Sensors.bInsideTemp = TRUE;
    g_Sensors.bBarom = TRUE;

    //// get prelim readings for rate initialization, where required
    i = 3;
    while( !GetBlock(DataBlock) )
    {
        i--;
        if( i == 0 )
        {
            g_log.PostMsg( HOSTMSG_GETBLOCK_FAIL, NULL );
            return FALSE;
        }
    }

    nVal = (short int) (DataBlock[TOUT] | (DataBlock[TOUT+1] << 8));
    if( nVal != WORD_ERROR )
    {
        nVal = FixValue( nVal );
        TempInitSet( nVal );
    }
    else
    {
        g_nErrTempOut++;
    }

    nVal = (short int) (DataBlock[BAR] | (DataBlock[BAR+1] << 8));
    if( nVal != UWORD_ERROR )
    {
        nVal = FixValue( nVal );
        nVal = MillibarsT( nVal );
        BaromInitSet( nVal );
    }
    else
    {
        g_nErrBarometer++;
    }

    nVal = (short int) DataBlock[HOUT];
    if( nVal != BYTE_ERROR )
    {
        if( nVal <= 100 )
            HumidSet( nVal );
    }
    else
    {
        g_nErrHumidityOut++;
    }

    // worker thread

    if( !StartLoop() )
        return FALSE;

    if( !DavPStartThread( bVerbose ) )
        return FALSE;

    if( g_bVerbose )
        fprintf( stdout, "Waiting for 3 data packets to complete initialization..." );

    while( g_nBlockCount < 3 )
        nanosleep( &wait, NULL );

    if( g_bVerbose )
        fprintf( stdout, "done\n" );

    return TRUE;
}


////////////////////////////////////////////////////////////////////
// GetAck - get command ACK.  Returns TRUE success, FALSE failure
//

BOOL GetAck( void )
{

    unsigned char c;

    if( ReadWxStnBytes( &c, 1 ) )
    {
        if( c == ACK )
            return TRUE;
        else
            return FALSE;
    }

    g_log.PostMsg( HOSTMSG_READWXSTNBYTES_FAIL, " GetAck()" );

    return FALSE;
}


//////////////////////////////////////////////////////////////////////////////
// ResetLink - used to insure link is in correct mode to receive commands and
// send data.  Flushes buffer in process.  If last character is '\r', returns
// TRUE, otherwise FALSE.
//

BOOL ResetLink( void )
{
    byte_t cBuffer[BLOCK_SIZE*2];
    int i;

    int nTries;
    timespec ts;
    char szLF[] = "\n";

    //////////////////////// setup timespec struct for nanosleep
    ts.tv_sec = 0;
    ts.tv_nsec = 50000000; // 50 mS
    nTries = 3;

    while( nTries )
    {
        FlushWxStnBuffer();

        // send "RESET"
        if( !WriteWxStn( szLF, 1 ) )
        {
            nTries--;
            continue;
        }

        nanosleep( &ts, NULL );

        if( !WriteWxStn( szLF, 1 ) )
        {
            nTries--;
            continue;
        }

        // read buffer until no more
        i = DrainWxStnBuffer( cBuffer, BLOCK_SIZE*2 );

        // check last two characters for LF/CR
        if( i < 2 )
        {
            nTries--;
            continue;
        }

        if( memcmp( "\n\r", &cBuffer[i-2], 2 ) == 0 )
            return TRUE;

        nTries--;
    }

    return FALSE;
}


//////////////////////////////////////////////////////////////////////////////
// start/stop worker thread that collects line of data from wx station
//

BOOL DavPStartThread( BOOL bVerbose )
{
    int nStatus;

    if( bVerbose )
        fprintf( stdout, "Starting worker thread..." );

    nStatus = pthread_create( &g_hThread, NULL, &DavPThread, NULL );

    if( nStatus == 0 )
    {
        if( bVerbose )
            fprintf( stdout, "done.\n" );
        return TRUE;
    }

    if( bVerbose )
        fprintf( stderr, "\n" );

    fprintf( stderr, "Failed to start worker thread.\n" );
    return FALSE;
}


void DavPStopThread( BOOL bVerbose )
{
    g_bStopThread = TRUE;

    if( bVerbose )
        fprintf( stdout, "attempting to stop worker thread..." );

    pthread_join( g_hThread, NULL );

    FlushWxStnBuffer();

    if( bVerbose )
        fprintf( stdout, "done.\n" );

    g_bStopThread = FALSE;
}


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// DavThread - thread that retrieves auto send data from the Davis station.
//

void * DavPThread( void * pParam )
{
    int nSpeed, nDir;
    int nTemp;
    int i;
    timespec wait;
    unsigned char c;
    byte_t DataBlock[BLOCK_SIZE];
    void * pBlackHole;

    pBlackHole = pParam;

    wait.tv_sec = 0;
    wait.tv_nsec = 5000;

    //FOREVER
    for(;;)
    {
        ////////////////////////// get data packet from station
        i = 0;

        //FOREVER     // for convenience of starting over after error condition
        for(;;)
        {

            if( g_bStopThread )
                pthread_exit( NULL );

            if( !ReadWxStnBytes( &c, 1 ) )
                continue;

            switch( i )
            {
                case 0:
                    if( c != 'L' )
                        continue;
                    break;

                case 1:
                    if( c != 'O' )
                    {
                        i=0;
                        continue;
                    }
                    break;

                case 2:
                    if( c != 'O' )
                    {
                        i=0;
                        continue;
                    }
                    break;
            }

            DataBlock[i++] = c;

            if( i < 3 )
                continue;

            do
            {
                if( g_bStopThread )
                    pthread_exit( NULL );

                if( !ReadWxStnBytes( &c, 1 ) )
                    continue;

                DataBlock[i++] = c;

            } while( i < BLOCK_SIZE );

            if( !CrcDav( DataBlock, BLOCK_SIZE) )
            {
                i = 0;
                g_nErrCrc++;
                continue;
            }
            break;      // block complete and verified
        }

        pthread_mutex_lock( &mx_DavPLock );

        memcpy( &g_DataBlock, &DataBlock, BLOCK_SIZE );

        g_nBlockCount++;

        ////////////// Wind

        nSpeed = (short int) g_DataBlock[WSPD];
        if( nSpeed != BYTE_ERROR )
        {
            // accumulate for average
            g_nSpeedCount++;
            g_nSpeedAccum += nSpeed;

            nDir = (short int) ( g_DataBlock[WDIR] | (g_DataBlock[WDIR+1] << 8) );
            if( nDir != WORD_ERROR )
            {
                g_nDirCount++;
                g_nDirAccum += nDir;
                // need to track quadrants to correct for QI/QIV ambiguity
                if( nDir < 90 )
                    g_nDaVPQ1++;
                else if ( nDir > 270 )
                    g_nDaVPQ4++;
            }
            else    // log sensor error
            {
                nDir = -1;
                g_nErrWindDir++;
            }

            if( nSpeed >= g_nGustSpeed )
            {
                g_nGustSpeed = nSpeed;
                g_nGustDir = nDir;
            }
        }
        else    // log sensor error
        {
            g_nErrWindSpd++;
        }


        ////////////// Temperature

        nTemp = (short int) ( g_DataBlock[TOUT] | (g_DataBlock[TOUT+1] << 8));

        if( nTemp != WORD_ERROR )
        {
            nTemp = FixValue( nTemp );
            if( nTemp >= g_nHiTemp )
                g_nHiTemp = nTemp;

            if( nTemp <= g_nLoTemp )
                g_nLoTemp = nTemp;
        }
        else
        {
            g_nErrTempOut++;
        }

        pthread_mutex_unlock( &mx_DavPLock );

        ////////////////////////// wait/check

        if( g_bStopThread )
            return NULL;

        nanosleep( &wait, NULL );

    }

    return NULL;
}


void DavPLog( void )
{
    char szString[32];

    if( g_nErrCrc )
    {
        sprintf( szString, "%i", g_nErrCrc );
        g_log.PostMsg( HOSTMSG_WXSTN_CRC_FAIL, szString );
        g_nErrCrc = 0;
    }

    if( g_nErrTempOut )
    {
        sprintf( szString, "%i", g_nErrTempOut );
        g_log.PostMsg( HOSTMSG_OUT_TEMP_READ_FAIL, szString );
        g_nErrTempOut = 0;
    }
    if( g_nErrTempIn )
    {
        sprintf( szString, "%i", g_nErrTempIn );
        g_log.PostMsg( HOSTMSG_IN_TEMP_READ_FAIL, szString );
        g_nErrTempIn = 0;
    }

    if( g_nErrBarometer )
    {
        sprintf( szString, "%i", g_nErrBarometer );
        g_log.PostMsg( HOSTMSG_BAROMETER_READ_FAIL, szString );
        g_nErrBarometer = 0;
    }

    if( g_nErrHumidityOut )
    {
        sprintf( szString, "%i", g_nErrHumidityOut );
        g_log.PostMsg( HOSTMSG_OUT_HUMIDITY_READ_FAIL, szString );
        g_nErrHumidityOut = 0;
    }
    if( g_nErrHumidityIn )
    {
        sprintf( szString, "%i", g_nErrHumidityIn );
        g_log.PostMsg( HOSTMSG_IN_HUMIDITY_READ_FAIL, szString );
        g_nErrHumidityIn = 0;
    }

    if( g_nErrPrecip )
    {
        sprintf( szString, "%i", g_nErrPrecip );
        g_log.PostMsg( HOSTMSG_PRECIP_READ_FAIL, szString );
        g_nErrPrecip = 0;
    }

}

