/***************************************************************************
 *   Copyright (C) 2005 by John Bennett   *
 *   n4xi@arrl.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

//
// id5.cpp - Heath ID-5001 weather station driver
//

#define _ID5_CPP_

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <pthread.h>
#include <unistd.h>

// libwxn
#include <logmsg.h>

#include "precip.h"
#include "barom.h"
#include "humid.h"
#include "temp.h"
#include "wind.h"
#include "wxport.h"
#include "sensors.h"
#include "rates.h"
#include "hglobal.h"
#include "wxdata.h"
#include "id5.h"


#define MAX_DATA_SIZE   128

//////////////////////////////////////
// local function prototypes
//////////////////////////////////////

BOOL GetId5Data( const char * psCmd, int nLen, char * psRsp );
int ExtractVal( char * pS, int nLen );
BOOL SendId5Cmd( const char * psCmd, int nLen );

// thread related functions

BOOL Id5StartThread( BOOL bVerbose );
void * Id5Line( void * pParam );


//////////////////////////////////////
// global variables @ file scope

static BOOL g_bStopThread = FALSE;

//////////////////////////////////////

pthread_mutex_t mx_Id5Lock = PTHREAD_MUTEX_INITIALIZER;

// worker thread
static pthread_t g_hThread;

// temporary values
static int g_nGustSpeed = 0;
static int g_nGustDir = 0;
static int g_nHiTemp;
static int g_nLoTemp;

// flags
static BOOL g_bWindRead = FALSE;
static BOOL g_bTempRead = FALSE;


////////////////////////////////////////////////
// ID5001 commands
//

const char * g_szInitCmd[] =
{
    "ATLS",     // send CR/LF
    "ATEC",     // no echo
    "ATXCA",    // stop auto send all
    "ATAC"      // alarm clear/off
};

#define INIT_CMD_CNT sizeof( g_szInitCmd ) / sizeof( g_szInitCmd[0] )

// all maintenance commands clear respective parameter
const char * g_szMaintCmd[] =
{
    "ATCBH",    // barom hi
    "ATCBL",    // barom lo
    "ATCHOH",   // humidity outdoor hi
    "ATCHOL",   // humidity outdoor lo
    "ATCTOH",   // temp outdoor hi
    "ATCTOL",   // temp outdoor lo
    "ATCWAH",   // wind avg hi
    "ATCWAL",   // wind avg lo
    "ATCWGH",   // wind gust hi
    "ATCWGL",   // wind gust lo
    "ATCR",     // precip
    "ATCHIH",   // humidity indoor hi
    "ATCHIL",   // humidity indoor lo
    "ATCTIH",   // temp indoor hi
    "ATCTIL"    // temp indoor lo
};

#define MAINT_CMD_CNT sizeof( g_szMaintCmd ) / sizeof( g_szMaintCmd[0] )


char g_szClrPrecipCmd[] = "ATCR";
#define CLR_PRECIP_CMD_LEN = sizeof( g_szClrPrecip ) - 1
char g_szDateCmd[] = "ATRD";
#define DATE_CMD_LEN sizeof( g_szDateCmd ) - 1
char g_szKickString[] = "AT";     // get an '8' response
#define KICK_CMD_LEN sizeof( g_szKickString ) - 1

// data commands

char g_szBcrCmd[] = "ATRB";      // barometer
#define BCR_CMD_LEN sizeof( g_szBcrCmd ) - 1
#define BCR_OFFSET  1
#define BCR_LEN     4

char g_szTcrCmd[] = "ATRTO";     // outdoor temperature
#define TCR_CMD_LEN sizeof( g_szTcrCmd ) - 1
#define TCR_OFFSET  1
#define TCR_LEN     3

char g_szIstCmd[] = "ATRTI";     // indoor temperature
#define IST_CMD_LEN sizeof( g_szIstCmd ) - 1
#define IST_OFFSET  1
#define IST_LEN     3

char g_szHcrCmd[] = "ATRHO";     // outdoor humidity
#define HCR_CMD_LEN sizeof( g_szHcrCmd ) - 1
#define HCR_OFFSET  1
#define HCR_LEN     2

char g_szIshCmd[] = "ATRHI";     // indoor humidity
#define ISH_CMD_LEN sizeof( g_szIshCmd ) - 1
#define ISH_OFFSET  1
#define ISH_LEN     2

char g_szWacCmd[] = "ATRWA";     // wind avg
#define WAC_CMD_LEN sizeof( g_szWacCmd ) - 1
#define WAC_SPD_OFFSET  1
#define WAC_SPD_LEN     3
#define WAC_DIR_OFFSET  5
#define WAC_DIR_LEN     3

char g_szWgcCmd[] = "ATRWG";     // wind gust
#define WGC_CMD_LEN sizeof( g_szWgcCmd ) - 1
#define WGC_SPD_OFFSET  1
#define WGC_SPD_LEN     3
#define WGC_DIR_OFFSET  5
#define WGC_DIR_LEN     3

char g_szPcaCmd[] = "ATRR";      // precip accum since midnight
#define PCA_CMD_LEN sizeof( g_szPcaCmd ) - 1
#define PCA_OFFSET  1
#define PCA_LEN     5

// auto data commands

char g_szAutoOnCmd[] = "ATXSTW";    // auto send temp, wind gust
#define AUTO_ON_CMD_LEN sizeof( g_szAutoOnCmd ) - 1
char g_szAutoOffCmd[] = "ATXCA\r";    // stop auto send all
#define AUTO_OFF_CMD_LEN sizeof( g_szAutoOffCmd ) - 1


//////////////////////////////////////////////////////////////////////////////
// Id5Reset - reset highs/lows/precip @ midnight
//

void Id5Reset( void )
{
    unsigned i;
    timespec wait;

    // This interval *MUST* be longer than the one in Id5Line(), otherwise
    // a deadlock may occur. Fair warning.
    wait.tv_sec = 0;
    wait.tv_nsec = 500000000;   // 500 mS

    if( !WriteWxStn( g_szAutoOffCmd, AUTO_OFF_CMD_LEN ) )
        return;

    pthread_mutex_lock( &mx_Id5Lock );

    for( i=0; i<MAINT_CMD_CNT; i++ )
    {
        if( !SendId5Cmd( g_szMaintCmd[i], strlen(g_szMaintCmd[i]) ) )
            g_log.PostMsg( HOSTMSG_SENDID5CMD_FAIL, NULL );
    }

    if( !SendId5Cmd( g_szAutoOnCmd, AUTO_ON_CMD_LEN ) )
        g_log.PostMsg( HOSTMSG_SENDID5CMD_FAIL, NULL );

    // allow worker thread to re-capture control;
    nanosleep( &wait, NULL );

    pthread_mutex_unlock( &mx_Id5Lock );

}


//////////////////////////////////////////////////////////////////////////////
// ProcessData - read weather station, set hi/low values, etc. Returns TRUE
// success, FALSE on failure.
//

BOOL Id5ProcessData( void )
{
    char szResp[MAX_DATA_SIZE+1];
    char szError[128];
    int nVal;
    cwinds Wind;

    if( !WriteWxStn( g_szAutoOffCmd, AUTO_OFF_CMD_LEN ) )
        return FALSE;

    pthread_mutex_lock( &mx_Id5Lock );

    ////////////////////////////////// TEMPERATURE
    if( !GetId5Data( g_szTcrCmd, TCR_CMD_LEN, szResp ) )
    {
        SendId5Cmd( g_szAutoOnCmd, AUTO_ON_CMD_LEN );
        pthread_mutex_unlock( &mx_Id5Lock );
        return FALSE;
    }
    if( *szResp == 'T' )
    {
        nVal = ExtractVal( &szResp[TCR_OFFSET], TCR_LEN );
        if( !g_bTempRead )
        {
            g_nHiTemp = nVal;
            g_nLoTemp = nVal;
        }

        TempSet( nVal, g_nHiTemp, g_nLoTemp );

        // reset hi/lo temps
        g_nHiTemp = nVal;
        g_nLoTemp = nVal;
    }
    else
    {
        sprintf( szError, "tcr-> [%s]", szResp );
        g_log.PostMsg( HOSTMSG_GETID5DATA_FAIL, szError );
    }
    g_bTempRead = FALSE;

    // inside


    if( !GetId5Data( g_szIstCmd, IST_CMD_LEN, szResp ) )
    {
        SendId5Cmd( g_szAutoOnCmd, AUTO_ON_CMD_LEN );
        pthread_mutex_unlock( &mx_Id5Lock );
        return FALSE;
    }

    if( *szResp == 't' )
    {
        nVal = ExtractVal( &szResp[IST_OFFSET], IST_LEN );
        InsideTempSet( nVal );
    }
    else
    {
        sprintf( szError, "ist-> [%s]", szResp );
        g_log.PostMsg( HOSTMSG_GETID5DATA_FAIL, szError );
    }

    ////////////////////////////////// WIND

    bzero( &Wind, sizeof( struct cwinds ) );

    //// average

    if( !GetId5Data( g_szWacCmd, WAC_CMD_LEN, szResp ) )
    {
        SendId5Cmd( g_szAutoOnCmd, AUTO_ON_CMD_LEN );
        pthread_mutex_unlock( &mx_Id5Lock );
        return FALSE;
    }

    if( *szResp == 'w' )
    {
        Wind.nSpeed = ExtractVal( &szResp[WAC_SPD_OFFSET], WAC_SPD_LEN );
        Wind.nDir = ExtractVal( &szResp[WAC_DIR_OFFSET], WAC_DIR_LEN );
    }
    else
    {
        Wind.nSpeed = 0;
        // Wind.nDir = -1;
        sprintf( szError, "wac-> [%s]", szResp );
        g_log.PostMsg( HOSTMSG_GETID5DATA_FAIL, szError );
    }

    //// gust

    if( !g_bWindRead )
    {
        if( !GetId5Data( g_szWgcCmd, WGC_CMD_LEN, szResp ) )
        {
            SendId5Cmd( g_szAutoOnCmd, AUTO_ON_CMD_LEN );
            pthread_mutex_unlock( &mx_Id5Lock );
            return FALSE;
        }

        if( *szResp == 'W' )
        {
            Wind.nGustSpeed = ExtractVal( &szResp[WGC_SPD_OFFSET], WGC_SPD_LEN );
            Wind.nGustDir = ExtractVal( &szResp[WGC_DIR_OFFSET], WGC_DIR_LEN );
        }
        else
        {
            Wind.nGustSpeed = 0;
            Wind.nGustDir = -1;
            sprintf( szError, "wgc-> [%s]", szResp );
            g_log.PostMsg( HOSTMSG_GETID5DATA_FAIL, szError );
         }
    }
    else
    {
        Wind.nGustSpeed = g_nGustSpeed;
        Wind.nGustDir = g_nGustDir;
    }

    // reset gust speed/dir/flags
    g_nGustSpeed = 0;
    g_nGustDir = 0;
    g_bWindRead = FALSE;

    WindSet( &Wind );

    ////////////////////////////////// Precip

    if( !GetId5Data( g_szPcaCmd, PCA_CMD_LEN, szResp ) )
    {
        SendId5Cmd( g_szAutoOnCmd, AUTO_ON_CMD_LEN );
        pthread_mutex_unlock( &mx_Id5Lock );
        return FALSE;
    }

    if( *szResp == 'R' )
    {
        nVal = ExtractVal( &szResp[PCA_OFFSET], PCA_LEN );
        PrecipSet( nVal );
    }
    else
    {
        sprintf( szError, "pca: [%s]", szResp );
        g_log.PostMsg( HOSTMSG_GETID5DATA_FAIL, szError );
    }

    ////////////////////////////////// Barometric Pressure

    if( !GetId5Data( g_szBcrCmd, BCR_CMD_LEN, szResp ) )
    {
        SendId5Cmd( g_szAutoOnCmd, AUTO_ON_CMD_LEN );
        pthread_mutex_unlock( &mx_Id5Lock );
        return FALSE;
    }

    if( *szResp == 'B' )
    {
        nVal = MillibarsT( ExtractVal( &szResp[BCR_OFFSET], BCR_LEN ) );
        BaromSet( nVal );
    }
    else
    {
        sprintf( szError, "bcr-> [%s]", szResp );
        g_log.PostMsg( HOSTMSG_GETID5DATA_FAIL, szError );
    }

    ////////////////////////////////// Humidity
    if( g_Sensors.bOutsideHumid )
    {
        if( !GetId5Data( g_szHcrCmd, HCR_CMD_LEN, szResp ) )
        {
            SendId5Cmd( g_szAutoOnCmd, AUTO_ON_CMD_LEN );
            pthread_mutex_unlock( &mx_Id5Lock );
            return FALSE;
        }
        if( *szResp == 'H' )
        {
            nVal = ExtractVal( &szResp[HCR_OFFSET], HCR_LEN );
            HumidSet( nVal );
        }
        else
        {
            sprintf( szError, "hcr-> [%s]", szResp );
            g_log.PostMsg( HOSTMSG_GETID5DATA_FAIL, szError );
        }
    }


    if( g_Sensors.bInsideHumid )
    {
        if( !GetId5Data( g_szIshCmd, ISH_CMD_LEN, szResp ) )
        {
            SendId5Cmd( g_szAutoOnCmd, AUTO_ON_CMD_LEN );
            pthread_mutex_unlock( &mx_Id5Lock );
            return FALSE;
        }
        if( *szResp == 'h' )
        {
            nVal = ExtractVal( &szResp[ISH_OFFSET], ISH_LEN );
            InsideHumidSet( nVal );
        }
        else
        {
            sprintf( szError, "ish-> [%s]", szResp );
            g_log.PostMsg( HOSTMSG_GETID5DATA_FAIL, szError );
        }
    }


    ////////////////////////////////// Clean-Up
    if( !SendId5Cmd( g_szAutoOnCmd, AUTO_ON_CMD_LEN ) )
        g_log.PostMsg( HOSTMSG_ID5RESTART_FAIL, NULL );

    pthread_mutex_unlock( &mx_Id5Lock );

    return TRUE;
}


//////////////////////////////////////////////////////////////////////////////
// GetId5Data - for use with commands that return data from ID5001. If
// time-out, zero length, or '8' rx'd, returns FALSE, otherwise TRUE.
//

BOOL GetId5Data( const char * psCmd, int nLen, char * psRsp )
{
    char szCmd[32];

    strcpy( szCmd, psCmd );
    strcat( szCmd, "\r" );

    if( !WriteWxStn( szCmd, nLen+1 ) )
        return FALSE;

    if( !ReadWxStn( psRsp, MAX_DATA_SIZE ) )
        return FALSE;

    if( strlen( psRsp ) == 0 )
        return FALSE;

    if( *psRsp == '8' )
        return FALSE;

    return TRUE;
}


//////////////////////////////////////////////////////////////////////////////
// SendId5Cmd - for use with commands that do *NOT* return data from ID5001. If
// time-out, zero length, or '8' rx'd, returns FALSE, otherwise TRUE.
//

BOOL SendId5Cmd( const char * psCmd, int nLen )
{
    char szResp[8];
    char szCmd[32];

    strcpy( szCmd, psCmd );
    strcat( szCmd, "\r" );

    if( !WriteWxStn( szCmd, nLen+1 ) )
        return FALSE;

    if( !ReadWxStn( szResp, 7 ) )
        return FALSE;

    if( *szResp == '0' )
        return TRUE;
    else
        return FALSE;
}


//////////////////////////////////////////////////////////////////////////////
// Id5Init - check connection, determine attached sensors.
//

BOOL Id5Init( BOOL bVerbose )
{
    unsigned i;
    int nVal;
    char szLine[MAX_DATA_SIZE+1];

    FlushWxStnBuffer();

    for( i=0; i<INIT_CMD_CNT; i++ )
    {
        if( !SendId5Cmd( g_szInitCmd[i], strlen(g_szInitCmd[i]) ) )
            return FALSE;
    }

    if( bVerbose )
        fprintf( stdout, "Requesting date from ID-5001: " );

    if( !GetId5Data( g_szDateCmd, DATE_CMD_LEN, szLine ) )
        return FALSE;

    if( bVerbose )
        fprintf( stdout, "%s\n", szLine );

    // outside humidity
    if( GetId5Data( g_szHcrCmd, HCR_CMD_LEN, szLine ) )
    {
        nVal = ExtractVal(&szLine[HCR_OFFSET], HCR_LEN);
        if( nVal )
        {
            HumidInitSet( nVal );
            g_Sensors.bOutsideHumid = TRUE;
            if( bVerbose )
                fprintf( stdout, "Found outside humidity sensor.\n" );
        }
    }

    // inside humidity
    if( GetId5Data( g_szIshCmd, ISH_CMD_LEN, szLine ) )
    {
        if( ExtractVal(&szLine[ISH_OFFSET], ISH_LEN) )
        {
            g_Sensors.bInsideHumid = TRUE;
            if( bVerbose )
                fprintf( stdout, "Found outside humidity sensor.\n" );
        }

    }

    // barometer
    g_Sensors.bBarom = TRUE;
    if( GetId5Data( g_szBcrCmd, BCR_CMD_LEN, szLine ) )
    {
        nVal = MillibarsT( ExtractVal(&szLine[BCR_OFFSET], BCR_LEN) );
        BaromInitSet( nVal );
    }

    // inside temperature
    g_Sensors.bInsideTemp = TRUE;

    ////////////// initialize variables used by thread

    // temp
    if( GetId5Data( g_szTcrCmd, TCR_CMD_LEN, szLine ) )
    {
        nVal = ExtractVal(&szLine[TCR_OFFSET], TCR_LEN);
        TempInitSet( nVal );
        g_nHiTemp = nVal;
        g_nLoTemp = nVal;
    }

    else
        return FALSE;

    // precip
    if( GetId5Data( g_szPcaCmd, PCA_CMD_LEN, szLine ) )
        PrecipSet( ExtractVal(&szLine[PCA_OFFSET], PCA_LEN) );
    else
        return FALSE;

    //////////////

    if( !SendId5Cmd( g_szAutoOnCmd, AUTO_ON_CMD_LEN) )
        return FALSE;

    if( !Id5StartThread( bVerbose ) )
        return FALSE;

    return TRUE;
}


//////////////////////////////////////////////////////////////////////////////
// ExtractVal - returns numeric value from string of n characters - position
// independent.
//

int ExtractVal( char * pS, int nLen )
{
    int nVal;
    char szString[16];

    memcpy( szString, pS, nLen );
    szString[nLen] = 0;

    nVal = atoi( szString );

    return nVal;
}


//////////////////////////////////////////////////////////////////////////////
// start/stop worker thread that collects line of data from wx station
//

BOOL Id5StartThread( BOOL bVerbose )
{
    int nStatus;

    if( bVerbose )
        fprintf( stdout, "Starting worker thread..." );

    nStatus = pthread_create( &g_hThread, NULL, &Id5Line, NULL );

    if( nStatus == 0 ) {
        if( bVerbose )
            fprintf( stdout, "done.\n" );
        return TRUE;
        }

    if( bVerbose )
        fprintf( stderr, "\n" );

    fprintf( stderr, "Failed to start worker thread.\n" );
    return FALSE;
}


void Id5StopThread( BOOL bVerbose )
{
    g_bStopThread = TRUE;

    if( bVerbose )
        fprintf( stdout, "attempting to stop worker thread..." );

    pthread_join( g_hThread, NULL );

    FlushWxStnBuffer();

    if( bVerbose )
        fprintf( stdout, "done.\n" );

    g_bStopThread = FALSE;
}


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// Id5Line - thread that retrieves auto send data from the ID5001.
//

void * Id5Line( void * pParam )
{
    timespec wait;
    char szLine[MAX_DATA_SIZE+1];
    int nSpeed;
    int nDir;
    int nTemp;
    void * pBlackHole;

    pBlackHole = pParam;

    // This interval *MUST* be shorter than the one in Id5Reset(), otherwise
    // a deadlock may occur. Fair warning.
    wait.tv_sec = 0;
    wait.tv_nsec = 100000000;   // 100 mS

    pthread_mutex_lock( &mx_Id5Lock );

    //FOREVER
    for(;;)
    {
        ////////////////////////////////// get line from station
        //FOREVER
        for(;;)
        {
            if( WxStnDataReady() )
            {
                if( ReadWxStn( szLine, STRING_SIZE ) )
                {
                    if( *szLine != '8' )
                        break;
                }
            }

            // catch a shutdown, if it occurred while reading or waiting
            if( g_bStopThread )
            {
                // must compensate for CR at end of command
                SendId5Cmd( g_szAutoOffCmd, AUTO_OFF_CMD_LEN-1 );
                pthread_mutex_unlock( &mx_Id5Lock );
                pthread_exit( NULL );
            }

        }

        ////////////////////////////////// Wind Gust
        if( *szLine == 'W' )
        {
            nSpeed = ExtractVal( &szLine[WGC_SPD_OFFSET], WGC_SPD_LEN );
            if( nSpeed )
            {
                nDir = ExtractVal( &szLine[WGC_DIR_OFFSET], WGC_DIR_LEN );
                if( nSpeed > g_nGustSpeed )
                {
                    g_nGustSpeed = nSpeed;
                    g_nGustDir = nDir;
                    g_bWindRead = TRUE;
                }
            }
        }


        ////////////////////////////////// Temperature

        else if( *szLine == 'T' )
        {
            g_bTempRead = TRUE;

            nTemp = ExtractVal( &szLine[TCR_OFFSET], TCR_LEN );

            // only highs/lows are checked - no current
            if( nTemp > g_nHiTemp )
                g_nHiTemp = nTemp;

            if( nTemp < g_nLoTemp )
                g_nLoTemp = nTemp;
        }

        ////////////////////////////////// ATXCA sent

        else if( *szLine == '0' )
        {
            // unlock this thread
            pthread_mutex_unlock( &mx_Id5Lock );

            // allow other thread to acquire control...
            nanosleep( &wait, NULL );

            // and wait to restart
            pthread_mutex_lock( &mx_Id5Lock );
        }

        // else ignore

    }

    return NULL;
}

