//---------------------------------------------------------------------------
// Copyright (C) 2000 Dallas Semiconductor Corporation, All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
// OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// Except as contained in this notice, the name of Dallas Semiconductor
// shall not be used except as stated in the Dallas Semiconductor
// Branding Policy.
//---------------------------------------------------------------------------
//
//  linuxlnk.C - COM functions required by MLANLL.C, MLANTRNU, MLANNETU.C and
//           MLanFile.C for MLANU to communicate with the DS2480 based
//           Universal Serial Adapter 'U'.  Fill in the platform specific code.
//
//  Version: 1.02
//
//  History: 1.00 -> 1.01  Added function msDelay.
//
//           1.01 -> 1.02  Changed to generic OpenCOM/CloseCOM for easier
//                         use with other platforms.
//

//--------------------------------------------------------------------------
// Copyright (C) 1998 Andrea Chambers and University of Newcastle upon Tyne,
// All Rights Reserved.
//--------------------------------------------------------------------------
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE UNIVERSITY OF NEWCASTLE UPON TYNE OR ANDREA CHAMBERS
// BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//---------------------------------------------------------------------------
//
//  LinuxLNK.C - COM functions required by MLANLLU.C, MLANTRNU.C, MLANNETU.C
//             and MLanFile.C for MLANU to communicate with the DS2480 based
//             Universal Serial Adapter 'U'.  Platform specific code.
//
//  Version: 2.01
//  History: 1.02 -> 1.03  modifications by David Smiczek
//                         Changed to use generic OpenCOM/CloseCOM
//                         Pass port name to OpenCOM instead of hard coded
//                         Changed msDelay to handle long delays
//                         Reformatted to look like 'TODO.C'
//                         Added #include "ds2480.h" to use constants.
//                         Added function SetBaudCOM()
//                         Added function msGettick()
//                         Removed delay from WriteCOM(), used tcdrain()
//                         Added wait for byte available with timeout using
//                          select() in ReadCOM()
//
//           1.03 -> 2.00  Support for multiple ports. Include "ownet.h". Use
//                         'uchar'.  Reorder functions. Provide correct
//                         return values to OpenCOM.  Replace 'makeraw' call.
//                         Should now be POSIX.
//           2.00 -> 2.01  Added support for owError library.
//

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <time.h>
#include <termios.h>
#include <errno.h>
#include <sys/time.h>
#include <memory.h>
#include <sys/poll.h>

#include "ds2480.h"
#include "ownet.h"

// LinuxLNK global
int g_fd;
struct termios g_origterm;

//---------------------------------------------------------------------------
// Attempt to open a com port.
// Set the starting baud rate to 9600.
//
// 'port_zstr' - zero terminate port name.  For this platform
//               use format COMX where X is the port number.
//
//
// Returns: TRUE(1)  - success, COM port opened
//          FALSE(0) - failure, could not open specified port
//

BOOL OpenCOM( char *port_zstr)
{
   struct termios t;               // see man termios - declared as above
   struct termios newterm;
   int rc;
   int tmp;


   g_fd = open(port_zstr, O_RDWR|O_NONBLOCK);
   if( g_fd<0 )
   {
      perror( port_zstr );
      OWERROR( OWERROR_GET_SYSTEM_RESOURCE_FAILED );
      return FALSE;
   }
   rc = tcgetattr( g_fd, &t );
   if( rc < 0 )
   {
      tmp = errno;
      perror( "one-wire" );
      close( g_fd );
      errno = tmp;
      OWERROR( OWERROR_SYSTEM_RESOURCE_INIT_FAILED );
      return FALSE;
   }

   // Save original settings.
   memcpy( &g_origterm, &t, sizeof( struct termios ) );

   bzero( &newterm, sizeof( newterm ) );

   // Set to non-canonical mode, and no RTS/CTS handshaking
   newterm.c_iflag = IGNBRK | IGNPAR;
   newterm.c_cflag = B9600 | CLOCAL | CS8 | CREAD;
   rc = tcsetattr( g_fd, TCSAFLUSH, &newterm );
   tcflush( g_fd,TCIOFLUSH );

   if( rc < 0 )
   {
      tmp = errno;
      perror( "one-wire" );
      close( g_fd );
      errno = tmp;
      OWERROR( OWERROR_SYSTEM_RESOURCE_INIT_FAILED );
      return FALSE;
   }

   return TRUE;
}


//---------------------------------------------------------------------------
// Closes the connection to the port.
//
//

void CloseCOM( void )
{

   // restore tty settings
   tcsetattr( g_fd, TCSAFLUSH, &g_origterm );
   FlushCOM();
   close( g_fd );
   g_fd = 0;

}


//--------------------------------------------------------------------------
// Write an array of bytes to the COM port, verify that it was
// sent out.  Assumes that baud rate has been set.
//
// Returns TRUE success, FALSE failure
//

BOOL WriteCOM( int outlen, uchar *outbuf)
{
    ssize_t count;

    count = write( g_fd, outbuf, outlen );
    tcdrain( g_fd );

    if( count == outlen )
        return TRUE;
    else
        return FALSE;
}


//--------------------------------------------------------------------------
// Read an array of bytes to the COM port, verify that it was
// sent out.  Assume that baud rate has been set.
//
// 'outlen'   - number of bytes to write to COM port
// 'outbuf'   - pointer ot an array of bytes to write
//
// Returns:  number of bytes read
//

int ReadCOM( int inlen, uchar *inbuf )
{
    struct pollfd Poll;
    int cnt;

    Poll.fd = g_fd;
    Poll.events = POLLIN | POLLPRI;

   // loop to wait until each byte is available and read it
    for( cnt=0; cnt<inlen; cnt++ )
    {
        // timeout is 20 mS
        if( poll( &Poll, 1, 20 ) == 0 )
            return cnt;     // timed out
        if( (Poll.revents | POLLIN) || (Poll.revents | POLLPRI) )
        {
            // if byte available read or return bytes read
            if( read( g_fd, &inbuf[cnt], 1 ) != 1 )
                return cnt;
        }
        else
            return cnt;
   }

   // success, so return desired length
   return inlen;

}


/*
int ReadCOM( int inlen, uchar *inbuf)
{
   fd_set         filedescr;
   struct timeval tval;
   int            cnt;


   // loop to wait until each byte is available and read it
   for( cnt = 0; cnt < inlen; cnt++ )
   {
      // set a descriptor to wait for a character available
      FD_ZERO( &filedescr );
      FD_SET( g_fd, &filedescr );
      // set timeout to 10ms
      tval.tv_sec = 0;
      tval.tv_usec = 10000;

      // if byte available read or return bytes read
      if( select(g_fd+1, &filedescr, NULL, NULL, &tval) != 0 )
      {
         if ( read( g_fd, &inbuf[cnt], 1 ) != 1 )
            return cnt;
      }
      else
         return cnt;
   }

   // success, so return desired length
   return inlen;
}
*/


//---------------------------------------------------------------------------
//  Description:
//     flush the rx and tx buffers
//
//

void FlushCOM( void )
{
   tcflush( g_fd, TCIOFLUSH );
}


//--------------------------------------------------------------------------
//  Description:
//     Send a break on the com port for at least 2 ms
//
//

void BreakCOM( void )
{
    int nDuration;

    nDuration = 0;                   // see man termios break may be too long
    tcsendbreak( g_fd, nDuration );

}


//--------------------------------------------------------------------------
// Set the baud rate on the com port.
//
// 'new_baud'  - new baud rate defined as
// PARMSET_9600     0x00
// PARMSET_19200    0x02
// PARMSET_57600    0x04
// PARMSET_115200   0x06
//

void SetBaudCOM( uchar new_baud )
{
    struct termios t;
    int rc;
    speed_t baud;

    // read the attribute structure
    rc = tcgetattr( g_fd, &t );
    if ( rc < 0 )
        return;

    // convert parameter to linux baud rate
    switch( new_baud )
    {
        case PARMSET_9600:
            baud = B9600;
            break;
        case PARMSET_19200:
            baud = B19200;
            break;
        case PARMSET_57600:
            baud = B57600;
            break;
        case PARMSET_115200:
            baud = B115200;
            break;
        default:
            baud = B9600;
            break;
   }

   // set baud in structure
   cfsetospeed( &t, baud );
   cfsetispeed( &t, baud );

   // change baud on port
   rc = tcsetattr( g_fd, TCSAFLUSH, &t );

}


//--------------------------------------------------------------------------
// Get the current millisecond tick count.  Does not have to represent
// an actual time, it just needs to be an incrementing timer.
//
long msGettick(void)
{
   struct timezone tmzone;
   struct timeval  tmval;
   long ms;

   gettimeofday(&tmval,&tmzone);
   ms = (tmval.tv_sec & 0xFFFF) * 1000 + tmval.tv_usec / 1000;
   return ms;

}


//--------------------------------------------------------------------------
//  Description:
//     Delay for at least 'len' ms
//
void msDelay(int len)

{
   struct timespec s;

   s.tv_sec = len / 1000;
   s.tv_nsec = (len - (s.tv_sec * 1000)) * 1000000;
   nanosleep( &s, NULL );
}

