//---------------------------------------------------------------------------
// Copyright (C) 2000 Dallas Semiconductor Corporation, All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
// OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// Except as contained in this notice, the name of Dallas Semiconductor
// shall not be used except as stated in the Dallas Semiconductor
// Branding Policy.
//---------------------------------------------------------------------------
//
//  owLLU.C - Link Layer 1-Wire Net functions using the DS2480/DS2480B (U)
//            serial interface chip.
//
//  Version: 2.01
//

#define _OWLLU_CPP_

#include "ownet.h"
#include "ds2480.h"


//--------------------------------------------------------------------------
// Reset all of the devices on the 1-Wire Net and return the result.
//
// Returns: TRUE(1):  presense pulse(s) detected, device(s) reset
//          FALSE(0): no presense pulses detected
//
//
int owTouchReset( void )
{
   uchar readbuffer[10],sendpacket[10];
   uchar sendlen=0;

   // make sure normal level
   owLevel( MODE_NORMAL );

   // check if correct mode
   if( g_nMode != MODSEL_COMMAND )
   {
      g_nMode = MODSEL_COMMAND;
      sendpacket[sendlen++] = MODE_COMMAND;
   }

   // construct the command
   sendpacket[sendlen++] = (uchar)(CMD_COMM | FUNCTSEL_RESET | g_nSpeed);

   // flush the buffers
   FlushCOM();

   // send the packet
   if( WriteCOM(sendlen,sendpacket) )
   {
      // read back the 1 byte response
      if( ReadCOM(1,readbuffer) == 1 )
      {
         // make sure this byte looks like a reset byte
         if( ((readbuffer[0] & RB_RESET_MASK) == RB_PRESENCE) || ((readbuffer[0] & RB_RESET_MASK) == RB_ALARMPRESENCE) )
            return TRUE;
         else
            OWERROR( OWERROR_RESET_FAILED );
      }
      else
         OWERROR( OWERROR_READCOM_FAILED );
   }
   else
      OWERROR( OWERROR_WRITECOM_FAILED );

   // an error occured so re-sync with DS2480
   DS2480Detect();

   return FALSE;
}

//--------------------------------------------------------------------------
// Send 1 bit of communication to the 1-Wire Net and return the
// result 1 bit read from the 1-Wire Net.  The parameter 'sendbit'
// least significant bit is used and the least significant bit
// of the result is the return bit.
//
// 'sendbit' - the least significant bit is the bit to send
//
// Returns: 0:   0 bit read from sendbit
//          1:   1 bit read from sendbit
//
int owTouchBit( int sendbit)
{
   uchar readbuffer[10],sendpacket[10];
   uchar sendlen=0;

   // make sure normal level
   owLevel(MODE_NORMAL);

   // check if correct mode
   if( g_nMode != MODSEL_COMMAND )
   {
      g_nMode = MODSEL_COMMAND;
      sendpacket[sendlen++] = MODE_COMMAND;
   }

   // construct the command
   sendpacket[sendlen] = (sendbit != 0) ? BITPOL_ONE : BITPOL_ZERO;
   sendpacket[sendlen++] |= CMD_COMM | FUNCTSEL_BIT | g_nSpeed;

   // flush the buffers
   FlushCOM();

   // send the packet
   if( WriteCOM(sendlen,sendpacket) )
   {
      // read back the response
      if( ReadCOM(1,readbuffer) == 1 )
      {
         // interpret the response
         if( ((readbuffer[0] & 0xE0) == 0x80) && ((readbuffer[0] & RB_BIT_MASK) == RB_BIT_ONE) )
            return 1;
         else
            return 0;
      }
      else
         OWERROR(OWERROR_READCOM_FAILED);
   }
   else
      OWERROR(OWERROR_WRITECOM_FAILED);

   // an error occured so re-sync with DS2480
   DS2480Detect();

   return 0;

}

//--------------------------------------------------------------------------
// Send 8 bits of communication to the 1-Wire Net and verify that the
// 8 bits read from the 1-Wire Net is the same (write operation).
// The parameter 'sendbyte' least significant 8 bits are used.
//
// 'sendbyte' - 8 bits to send (least significant byte)
//
// Returns:  TRUE: bytes written and echo was the same
//           FALSE: echo was not the same
//

int owWriteByte( int sendbyte)
{
   return (owTouchByte(sendbyte) == (0xff & sendbyte)) ? TRUE : FALSE;
}


//--------------------------------------------------------------------------
// Send 8 bits of read communication to the 1-Wire Net and and return the
// result 8 bits read from the 1-Wire Net.
//
// Returns:  8 bits read from 1-Wire Net
//
int owReadByte( void )
{
   return owTouchByte((int)0xFF);
}


//--------------------------------------------------------------------------
// Send 8 bits of communication to the 1-Wire Net and return the
// result 8 bits read from the 1-Wire Net.  The parameter 'sendbyte'
// least significant 8 bits are used and the least significant 8 bits
// of the result is the return byte.
//
// 'sendbyte' - 8 bits to send (least significant byte)
//
// Returns:  8 bits read from sendbyte
//
int owTouchByte( int sendbyte)
{
   uchar readbuffer[10],sendpacket[10];
   uchar sendlen=0;

   // make sure normal level
   owLevel(MODE_NORMAL);

   // check if correct mode
   if (g_nMode != MODSEL_DATA)
   {
      g_nMode = MODSEL_DATA;
      sendpacket[sendlen++] = MODE_DATA;
   }

   // add the byte to send
   sendpacket[sendlen++] = (uchar)sendbyte;

   // check for duplication of data that looks like COMMAND mode
   if (sendbyte ==(int)MODE_COMMAND)
      sendpacket[sendlen++] = (uchar)sendbyte;

   // flush the buffers
   FlushCOM();

   // send the packet
   if( WriteCOM(sendlen,sendpacket) )
   {
      // read back the 1 byte response
      if( ReadCOM(1,readbuffer) == 1 )
      {
          // return the response
          return (int)readbuffer[0];
      }
      else
         OWERROR(OWERROR_READCOM_FAILED);
   }
   else
      OWERROR(OWERROR_WRITECOM_FAILED);

   // an error occured so re-sync with DS2480
   DS2480Detect();

   return 0;

}


//--------------------------------------------------------------------------
// Set the 1-Wire Net line level.  The values for new_level are
// as follows:
//
// 'new_level' - new level defined as
//                MODE_NORMAL     0x00
//                MODE_STRONG5    0x02
//                MODE_PROGRAM    0x04
//                MODE_BREAK      0x08 (not supported)
//
// Returns:  current 1-Wire Net level
//

int owLevel( int new_level)
{
   uchar sendpacket[10],readbuffer[10];
   uchar sendlen=0;
   BOOL bStatus;
   BOOL bDoCheck;

   bDoCheck = FALSE;
   bStatus = FALSE;

   // check if need to change level
   if( new_level != g_nLevel )
   {
      // check if just putting back to normal
      if( new_level == MODE_NORMAL )
      {
         // check for disable strong pullup step
         if( g_nLevel == MODE_STRONG5 )
            bDoCheck = TRUE;

         // check if correct mode
         if( g_nMode != MODSEL_COMMAND )
         {
            g_nMode = MODSEL_COMMAND;
            sendpacket[sendlen++] = MODE_COMMAND;
         }

         // stop pulse command
         sendpacket[sendlen++] = MODE_STOP_PULSE;

         // flush the buffers
         FlushCOM();

         // send the packet
         if( WriteCOM(sendlen,sendpacket) )
         {
            msDelay( 4 );
            // read back the 1 byte response
            if( ReadCOM(1,readbuffer) == 1 )
            {
               // check response byte
               if ((readbuffer[0] & 0xE0) == 0xE0)
               {
                   bStatus = TRUE;
                   g_nLevel = MODE_NORMAL;
               }
               else
                   OWERROR( OWERROR_OWLEVEL_FAIL );
            }
            else
               OWERROR(OWERROR_READCOM_FAILED);
         }
         else
            OWERROR(OWERROR_WRITECOM_FAILED);

         // do extra bit for DS2480 disable strong pullup
         if( bDoCheck )
            owTouchBit(1);
      }
      // set new level
      else
      {
         // check if correct mode
         if( g_nMode != MODSEL_COMMAND )
         {
            g_nMode = MODSEL_COMMAND;
            sendpacket[sendlen++] = MODE_COMMAND;
         }

         // strong 5 volts
         if( new_level == MODE_STRONG5 )
         {
            // set the SPUD time value
            sendpacket[sendlen++] = CMD_CONFIG | PARMSEL_5VPULSE | PARMSET_infinite;
            // add the command to begin the pulse
            sendpacket[sendlen++] = CMD_COMM | FUNCTSEL_CHMOD | SPEEDSEL_PULSE | BITPOL_5V;
         }

         // flush the buffers
         FlushCOM();

         // send the packet
         if( WriteCOM(sendlen,sendpacket) )
         {
            // read back the 1 byte response from setting time limit
            if( ReadCOM(1,readbuffer) == 1 )
            {
               // check response byte
               if( (readbuffer[0] & 0x81) == 0 )
               {
                  g_nLevel = new_level;
                  bStatus = TRUE;
               }
               else
                   OWERROR( OWERROR_OWLEVEL_FAIL );
            }
            else
               OWERROR( OWERROR_READCOM_FAILED );
         }
         else
            OWERROR( OWERROR_WRITECOM_FAILED );
      }

      // if lost communication with DS2480 then reset
      if( !bStatus )
         DS2480Detect();
   }

   // return the current level
   return g_nLevel;
}


//--------------------------------------------------------------------------
// Send 8 bits of communication to the 1-Wire Net and verify that the
// 8 bits read from the 1-Wire Net is the same (write operation).
// The parameter 'sendbyte' least significant 8 bits are used.  After the
// 8 bits are sent change the level of the 1-Wire net.
//
// 'sendbyte' - 8 bits to send (least significant bit)
//
// Returns:  TRUE: bytes written and echo was the same, strong pullup now on
//           FALSE: echo was not the same
//

BOOL owWriteBytePower( int sendbyte )
{
   uchar sendpacket[10],readbuffer[10];
   uchar sendlen=0;
   BOOL bStatus;
   uchar i, temp_byte;

   // check if correct mode
   if (g_nMode != MODSEL_COMMAND)
   {
      g_nMode = MODSEL_COMMAND;
      sendpacket[sendlen++] = MODE_COMMAND;
   }

   // set the SPUD time value
   sendpacket[sendlen++] = CMD_CONFIG | PARMSEL_5VPULSE | PARMSET_infinite;

   // construct the stream to include 8 bit commands with the last one
   // enabling the strong-pullup
   temp_byte = sendbyte;
   for (i = 0; i < 8; i++)
   {
      sendpacket[sendlen++] = ((temp_byte & 0x01) ? BITPOL_ONE : BITPOL_ZERO)
                              | CMD_COMM | FUNCTSEL_BIT | g_nSpeed |
                              ((i == 7) ? PRIME5V_TRUE : PRIME5V_FALSE);
      temp_byte >>= 1;
   }

   // flush the buffers
   FlushCOM();

   bStatus = FALSE;

   // send the packet
   if( WriteCOM(sendlen,sendpacket) )
   {
      // read back the 9 byte response from setting time limit
      if( ReadCOM(9,readbuffer) == 9 )
      {
         // check response
         if( (readbuffer[0] & 0x81) == 0 )
         {
            // indicate the port is now at power delivery
            g_nLevel = MODE_STRONG5;

            // reconstruct the echo byte
            temp_byte = 0;
            for( i=0; i<8; i++ )
            {
               temp_byte >>= 1;
               temp_byte |= (readbuffer[i + 1] & 0x01) ? 0x80 : 0;
            }

            if( temp_byte == sendbyte )
               bStatus = TRUE;
            else
                OWERROR( OWERROR_OWWRITEBYTEPOWER_FAIL );
         }
      }
      else
         OWERROR(OWERROR_READCOM_FAILED);
   }
   else
      OWERROR(OWERROR_WRITECOM_FAILED);

   // if lost communication with DS2480 then reset
   if( !bStatus )
      DS2480Detect();

   return bStatus;

}


//--------------------------------------------------------------------------
// Send 1 bit of communication to the 1-Wire Net and verify that the
// response matches the 'applyPowerResponse' bit and apply power delivery
// to the 1-Wire net.  Note that some implementations may apply the power
// first and then turn it off if the response is incorrect.
//
// 'applyPowerResponse' - 1 bit response to check, if correct then start
//                        power delivery
//
// Returns:  TRUE: bit written and response correct, strong pullup now on
//           FALSE: response incorrect
//
int owReadBitPower( int applyPowerResponse)
{
   uchar sendpacket[3],readbuffer[3];
   uchar sendlen=0;
   uchar rt=FALSE;

   // check if correct mode
   if (g_nMode != MODSEL_COMMAND)
   {
      g_nMode = MODSEL_COMMAND;
      sendpacket[sendlen++] = MODE_COMMAND;
   }

   // set the SPUD time value
   sendpacket[sendlen++] = CMD_CONFIG | PARMSEL_5VPULSE | PARMSET_infinite;

   // enabling the strong-pullup after bit
   sendpacket[sendlen++] = BITPOL_ONE | CMD_COMM | FUNCTSEL_BIT | g_nSpeed | PRIME5V_TRUE;
   // flush the buffers
   FlushCOM();

   // send the packet
   if (WriteCOM(sendlen,sendpacket))
   {
      // read back the 2 byte response from setting time limit
      if (ReadCOM(2,readbuffer) == 2)
      {
         // check response to duration set
         if ((readbuffer[0] & 0x81) == 0)
         {
            // indicate the port is now at power delivery
            g_nLevel = MODE_STRONG5;

            // check the response bit
            if ((readbuffer[1] & 0x01) == applyPowerResponse)
               rt = TRUE;
            else
               owLevel(MODE_NORMAL);

            return rt;
         }
      }
      else
         OWERROR(OWERROR_READCOM_FAILED);
   }
   else
      OWERROR(OWERROR_WRITECOM_FAILED);

   // if lost communication with DS2480 then reset
   if (rt != TRUE)
      DS2480Detect();

   return rt;
}


