/***************************************************************************
 *   Copyright (C) 2005 by John Bennett   *
 *   n4xi@arrl.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#define _RAD_CPP_

#include <stdlib.h>
#include <libpq-fe.h>

// libwxn
#include <tz.h>

#include "wxdata.h"
#include "onewire.h"
#include "hconfig.h"
#include "rad.h"


static unsigned long g_ulRadLast = 0;
static unsigned long g_ulRadConvert;
static int g_nRadReadError = 0;


//////////////////////////////////////////////////////////////////////////////
// RadRead

void RadRead( void )
{
    unsigned long ulAccum;
    int nResult;

    if( !OwReadCounter( RAD_CNT, ulAccum ) )
    {
        g_nRadReadError++;
        g_Local.Data.rca = 0;
        return;
    }

    nResult = RadFix( ulAccum );
    if( g_nRadReadError )
    {
        g_Local.Data.rca = nResult / g_nRadReadError;
        g_nRadReadError = 0;
    }
    else
        g_Local.Data.rca = nResult;
}


//////////////////////////////////////////////////////////////////////////////
// RadFix - used to determine current rad count which should be the difference
// between the current and the last. If counter has been reset, current accum
// becomes current reading. The result is divided by five to give the the
// 'one-minute' rate. Writes accum to backend and returns current reading.
//

int RadFix( unsigned long ulVal )
{
    unsigned long ulCurrent;
    unsigned long ulResult;

    if( ulVal < g_ulRadLast )
        ulCurrent = ulVal;
    else
        ulCurrent = (ulVal - g_ulRadLast);
    g_ulRadLast = ulVal;

    // normalize data and correct to one minute intervals
    ulResult = (( ulCurrent * 100 )/g_ulRadConvert)/5;

    return (int) ulResult;
}


//////////////////////////////////////////////////////////////////////////////
// RadInit - initializes all rad values/structures. 'nCesium137' is the
// conversion standard for the radiation sensor used and is 1050 for the RM60
// unless there is an entry in <wxnhostd.conf> to override.
//

BOOL RadInit( int nCesium137 )
{
    unsigned long ulAccum;

    g_ulRadConvert = (unsigned long) nCesium137/10;

    if( !OwReadCounter( RAD_CNT, ulAccum ) )
    {
        g_ulRadLast = 0;
        return FALSE;
    }

    g_ulRadLast = ulAccum;

    return TRUE;
}


//////////////////////////////////////////////////////////////////////////////
// RadCheck - checks value to see if it exceeds alert value. If so, increments
// counter and sets nmax if appropriate.
//

void RadCheck( void )

{
    if( g_Local.Data.rca > g_HostConfig.Alarm.nHiRad )
        g_Local.Data.ralarm_count++;

    if( g_Local.Data.rca > g_Local.Data.rmax )
        g_Local.Data.rmax = g_Local.Data.rca;
}


//////////////////////////////////////////////////////////////////////////////
// RadAvg - returns average micro-roentgens over the last n hours
//

void RadAvg( void )
{
    int nMin;
    PGconn * pConn;
    PGresult * pResult;
    char szExec[128];

    pConn = PQconnectdb( "dbname=wxn" );
    if( pConn == NULL )
        return;

    if( PQstatus( pConn ) == CONNECTION_OK )
    {
        // force minute to be multiple of 5
        nMin = (g_Time.Minute()/5) * 5;

        // save current reading
        sprintf( szExec, "UPDATE rad SET rlt=%i WHERE hour=%i AND min=%i",
            g_Local.Data.rca, g_Time.Hour(), nMin );
        pResult = PQexec( pConn, szExec );
        PQclear( pResult );

        // hourly avg
        sprintf( szExec, "SELECT avg(rlt) FROM rad WHERE (hour=%i AND min>%i) OR (hour=%i AND min<=%i)",
            (g_Time.Hour()+23)%24, nMin, g_Time.Hour(), nMin );
        pResult = PQexec( pConn, szExec );
        if( PQresultStatus( pResult ) == PGRES_TUPLES_OK )
        {
            if( PQntuples( pResult ) )
                g_Local.Data.rha = atoi( PQgetvalue( pResult, 0, 0 ) );
            else
                g_Local.Data.rha = 0;
        }
        PQclear( pResult );

        // 24 hour avg
        sprintf( szExec, "SELECT avg(rlt) FROM rad" );
        pResult = PQexec( pConn, szExec );
        if( PQresultStatus( pResult ) == PGRES_TUPLES_OK )
        {
            if( PQntuples( pResult ) )
                g_Local.Data.rda = atoi( PQgetvalue( pResult, 0, 0 ) );
            else
                g_Local.Data.rda = 0;
        }
        PQclear( pResult );

        PQfinish( pConn );
    }
}


//////////////////////////////////////////////////////////////////////////////
// RadClear - resets daily max count to zero and rmax to current
//

void RadClear( void )

{
    g_Local.Data.ralarm_count = 0;
    g_Local.Data.rmax = g_Local.Data.rca;
}

