/***************************************************************************
 *   Copyright (C) 2005 by John Bennett   *
 *   n4xi@arrl.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

//
// umx.cpp - driver for all Ultimeter weather stations as well as TAPR
// T-238/T-238+ in Peet mode.
//

#define _UMX_CPP_

#include <string.h>
#include <stdio.h>
#include <time.h>
#include <pthread.h>
#include <unistd.h>

// libwxn
#include <htoi.h>

#include "precip.h"
#include "barom.h"
#include "humid.h"
#include "temp.h"
#include "wind.h"
#include "wxport.h"
#include "sensors.h"
#include "rates.h"
#include "wxdata.h"
#include "fixvalue.h"
#include "hconfig.h"
#include "umx.h"


//////////////////////////////////////
// local function prototypes
//////////////////////////////////////

int KphToMph( int nKph );
int KphToMph2( int nKph );

BOOL Check( char * szLine, int nLength );
int Compass( int nRawValue );
int Compass2( int nRawVal );

// thread related functions

BOOL UmxStartThread( int nType, BOOL bVerbose );
void * Um2Line( void * pParam );
void * U2kLine( void * pParam );

//////////////////////////////////////
// global variables @ file scope
//////////////////////////////////////

static char g_szInitCmd[] = ">I";

#define UINIT_CMD_SIZE sizeof( g_szInitCmd )

static BOOL g_bStopThread = FALSE;
pthread_mutex_t mx_UmxLock = PTHREAD_MUTEX_INITIALIZER;

// worker thread
static pthread_t g_hThread;

// weather instrument
static unsigned g_uUmxLength;
static int g_nUmxType;
static BOOL g_bKPH = FALSE;     // only for II series

// counters
static int g_nReadCount = 0;
static int g_nOverflowCount = 0;

// wind
static int g_nWindCount = 0;
static int g_nRawWindSpeedAccum = 0;
static int g_nRawWindDirAccum = 0;
static int g_nRawGustSpeed = 0;
static int g_nRawGustDirection = 0;
static int g_nQ1 = 0;
static int g_nQ4 = 0;

static int g_nPrecip = 0;
static int g_nBarom;

static int g_nOutHumid;
static int g_nInHumid;

static int g_nTemp;
static int g_nTempHi;
static int g_nTempLo;

static int g_nInsideTemp;


//////////////////////////////////////////////////////////////////////////////
// ProcessData - perform conversions, set hi/low values, etc. Returns TRUE if
// able to process, FALSE if not.
//

BOOL UmxProcessData( void )
{

    int nRawAvgSpd;
    int nRawAvgDir;
    cwinds Wind;

    memset( &Wind, 0, sizeof( cwinds ) );

    if( g_nReadCount == 0 )
        return FALSE;

    pthread_mutex_lock( &mx_UmxLock );

    ////////////////////////////////// TEMPERATURE

    TempSet( g_nTemp, g_nTempHi, g_nTempLo );

    // reset for next round
    g_nTempHi = g_nTemp;
    g_nTempLo = g_nTemp;

    ////////////////////////////////// Inside Temperature

    InsideTempSet( g_nInsideTemp );

    ////////////////////////////////// WIND

    bzero( &Wind, sizeof( struct cwinds ) );

    if( g_nWindCount )
    {
        //// averages

        nRawAvgSpd = g_nRawWindSpeedAccum / g_nWindCount;
        if( g_nUmxType == ULTII )
        {
            if( g_bKPH )
                Wind.nSpeed = KphToMph2( nRawAvgSpd );
            else
                Wind.nSpeed = nRawAvgSpd;
        }
        else    // Ultimeter 2000 series
            Wind.nSpeed = KphToMph( nRawAvgSpd );

        if( g_nQ1 && g_nQ4 )
        {
            if( g_nUmxType == ULTII )
                g_nRawWindDirAccum += g_nQ1 * UM2_360;
            else
                g_nRawWindDirAccum += g_nQ1 * U2K_360;
        }
        nRawAvgDir = g_nRawWindDirAccum / g_nWindCount ;
        if( g_nUmxType == ULTII )
            Wind.nDir = Compass2( nRawAvgDir );
        else    // Ultimeter 2000 series
            Wind.nDir = Compass( nRawAvgDir );
        // correct for averages over 359
        if( Wind.nDir > 359 )
            Wind.nDir -= 360;

        //// gusts

        if( g_nUmxType == ULTII )
        {
            if( g_bKPH )
                Wind.nGustSpeed = KphToMph2( g_nRawGustSpeed );
            else
                Wind.nGustSpeed = g_nRawGustSpeed;

            Wind.nGustDir = Compass2( g_nRawGustDirection );
        }
        else    // Ultimeter 2000 series
        {
            Wind.nGustSpeed = KphToMph( g_nRawGustSpeed );
            Wind.nGustDir = Compass( g_nRawGustDirection );
        }

        g_nQ1 = 0;
        g_nQ4 = 0;
        g_nWindCount = 0;
        g_nRawWindSpeedAccum = 0;
        g_nRawWindDirAccum = 0;
        g_nRawGustSpeed = 0;
        g_nRawGustDirection = 0;

    }
    else
        Wind.nGustDir = -1;

    WindSet( &Wind );


    ////////////////////////////////// Precip

    PrecipSet( g_nPrecip );

    ////////////////////////////////// Barometric Pressure

    if( g_Sensors.bBarom )
        BaromSet( g_nBarom );

    ////////////////////////////////// Humidity

    if( g_Sensors.bOutsideHumid )
        HumidSet( g_nOutHumid );

    if( g_Sensors.bInsideHumid )
        InsideHumidSet( g_nInHumid );

    ////////////////////////////////// Clean-Up

    // reset counters
    g_nReadCount = 0;
    g_nOverflowCount = 0;

    pthread_mutex_unlock( &mx_UmxLock );

    return TRUE;

}


//////////////////////////////////////////////////////////////////////////////
// UmxDataReady - returns true if station can be read, FALSE otherwise
//

BOOL UmxDataReady( void )
{

    if( g_nReadCount )
        return TRUE;
    else
        return FALSE;
}


//////////////////////////////////////////////////////////////////////////////
// Compass - returns degree reading in range of 0 to 360 from raw data input
// of 0-255. Returns 0 if input is greater than 255.
//

#define MAXDIR 255

int Compass( int nRawValue )
{

    if( nRawValue > MAXDIR )
        return 0;

    return (nRawValue * 141) / 100;

}


//////////////////////////////////////////////////////////////////////////////
// Compass2 - Converts compass direction as a single hex digit in the range
// 0 to F to a corresponding numeric value and returns that value.
//

int Compass2( int nRawVal )
{

    if( nRawVal > MAXDIR )
        return 0;

    return (nRawVal*225) / 10;

}


//////////////////////////////////////////////////////////////////////////////
// KphToMph - converts kph in 0.1 kph increments to mph (rounded).
// MPH = KPH*5/8.  Since values are in 0.1 KPH, result must be divided by
// 10 to get correct MPH.  Resulting formula is  MPH = KPH/16.  Rounding is
// performed as appropriate. Used for 2000, 800, 500, and 100 series.
//

int KphToMph( int nKph )
{

    if( nKph%16 > 7 )
        return nKph/16 + 1;
    else
        return nKph/16;

}


//////////////////////////////////////////////////////////////////////////////
// KphToMph2 - converts kph mph (rounded). MPH = KPH*5/8.  Rounding is
// performed as appropriate. Used for II series.
//

int KphToMph2( int nKph )
{

    int nVal;

    nVal = nKph * 5;

    if( nVal%8 > 4 )
        return nVal/8 + 1;
    else
        return nVal/8;

}



//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// Initialization Section
//


//////////////////////////////////////////////////////////////////////////////
// UmxInit - check connection, determine attached sensors.
//

BOOL UmxInit( BOOL bVerbose )
{

    char szLine[STRING_SIZE];

    FlushWxStnBuffer();

    if( g_nUmxType != ULTII )
        WriteWxStn( g_szInitCmd, UINIT_CMD_SIZE );

    if( bVerbose )
        fprintf( stdout, "Checking weather station connection:\n" );

    if( !Check( szLine, STRING_SIZE ) )
    {
        fprintf( stderr, "Timeout: wx station did not respond\n" );
        return FALSE;
    }
    else
    {
        if( bVerbose )
            fprintf( stdout, "Response: %s\n", szLine );
    }

    if( g_nUmxType != ULTII )
    {
        if( szLine[OHUM_OFF] != '-' )
        {
            g_Sensors.bOutsideHumid = TRUE;
            if( bVerbose )
                fprintf( stdout, "Found outside humidity sensor.\n" );
        }

        if( szLine[IHUM_OFF] != '-' )
        {
            g_Sensors.bInsideHumid = TRUE;
            if( bVerbose )
                fprintf( stdout, "Found inside humidity sensor.\n" );
        }

        if( szLine[ITMP_OFF] != '-' )
        {
            g_Sensors.bInsideTemp = TRUE;
            if( bVerbose )
                fprintf( stdout, "Found inside temperature sensor.\n" );
        }

        if( szLine[RAIN_OFF] != '-' )
        {
            g_Sensors.bRainGauge = TRUE;
            if( bVerbose )
                fprintf( stdout, "Found rain gauge.\n" );
        }

        if( szLine[BARM_OFF] != '-' )
        {
            g_Sensors.bBarom = TRUE;
            if( bVerbose )
                fprintf( stdout, "Found barometric pressure sensor.\n" );
        }
    }
    else    // Ultimeter II
    {
        g_Sensors.bRainGauge = TRUE;
    }

    if( g_Sensors.bRainGauge )
    {
        if( g_nUmxType != ULTII )
            g_nPrecip = htoi( &szLine[RAIN_OFF], RL4 );
        else
            g_nPrecip = htoi( &szLine[UM2_PRECIP], RL4 );

        if( g_HostConfig.nGaugeType == TENTHS )
            g_nPrecip *= 10;             // 0.1" gauge

        SetLastPrecip( g_nPrecip );
    }

    if( !UmxStartThread( g_nUmxType, bVerbose ) )
        return FALSE;

    if( bVerbose )
    {
        fprintf( stdout, "Waiting for five data records to determine initial wind values..." );
        fflush( stdout );
    }

    while( g_nReadCount < 5 )
        sleep( 1 );  // spin

    //// populate required fields for rate init

    TempInitSet( g_nTemp );
    if( g_Sensors.bOutsideHumid )
        HumidInitSet( g_nOutHumid );
    if( g_Sensors.bBarom )
        BaromInitSet( g_nBarom );

    if( bVerbose )
        fprintf( stdout, "done.\n" );

    return TRUE;

}


//////////////////////////////////////////////////////////////////////////////
// Check - get initial string from Ultimeter II, 100, 500, 800 or 2000
// series stations.  The algorthim determines the length of a data packet for
// that particular station and stores it at <g_nUmxLength> for use by the
// receiving routine. Returns TRUE success, FALSE failure.
//

BOOL Check( char * szLine, int nMaxLen )
{

    int nTries = 3;
    int nLength;
    BOOL bFound = FALSE;

    g_uUmxLength = 0;
    nLength = 0;

    while( nTries-- )
    {
        if( !ReadWxStn( szLine, nMaxLen ) )
            return FALSE;
        switch( *szLine )
        {
            case '*':   // early Ultimeter II
            case '#':   // late Ultimeter II
            case '!':   // Ultimeter 100, 200, 500, 800, and 2000 series
                // start of legit line found
                break;

            default:
                continue;
        }

        nLength = strlen( szLine );
        switch( nLength )
        {
            case U2KLEN1:
            case U2KLEN2:
            case U2KLEN3:
                g_nUmxType = ULT2000;
                bFound = TRUE;
                break;

            case UM2CNT:
                g_nUmxType = ULTII;
                bFound = TRUE;
                break;

            default:
                break;

        }

        if( bFound )
            break;
    };

    if( nLength )
    {
        g_uUmxLength = nLength;
        return TRUE;
    }

    // line had zero length or did not match anything
    return FALSE;

}


//////////////////////////////////////////////////////////////////////////////
// start/stop worker thread that collects line of data from wx station
//

BOOL UmxStartThread( int nType, BOOL bVerbose )
{
    int nStatus;

    if( bVerbose )
        fprintf( stdout, "Starting worker thread..." );
    if( nType != ULTII )
        nStatus = pthread_create( &g_hThread, NULL, &U2kLine, NULL );
    else
        nStatus = pthread_create( &g_hThread, NULL, &Um2Line, NULL );

    if( nStatus == 0 )
    {
        if( bVerbose )
            fprintf( stdout, "done.\n" );
        return TRUE;
    }

    fprintf( stderr, "Failed to start worker thread.\n" );
    return FALSE;

}


void UmxStopThread( BOOL bVerbose )
{

    g_bStopThread = TRUE;
    pthread_join( g_hThread, NULL );
    g_bStopThread = FALSE;

    if( bVerbose )
        fprintf( stdout, "umx driver thread halted\n" );

}


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// The secton that follows contains the threads that actually gather data
// from the Ultimeter 2000 and Ultimeter II series.
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// U2kLine - retrieves a complete line from the Ultimeter 2000 series
// stations. Converts raw values and stores. All data conversion takes place
// except calculation of average wind values which must be done every 'n'
// minutes.
//

void * U2kLine( void * pParam )
{

    char szLine[STRING_SIZE+1];
    int nRawSpeed;
    int nRawDirection;
    int nVal;
    int nTemp;
    void * pBlackHole;

    pBlackHole = pParam;

    //FOREVER
    for(;;)
    {

        ////////////////////////////////// get line from station

        //FOREVER
        for(;;)
        {

            if( g_bStopThread )
                pthread_exit( NULL );

            if( !ReadWxStn( szLine, STRING_SIZE ) )
                continue;

            if( strlen( szLine ) != g_uUmxLength )
                continue;

            // line must begin with this or something is wrong
            if( memcmp( szLine, "!!", 2 ) )
                continue;
            else
                break;      // legit line

        }

        pthread_mutex_lock( &mx_UmxLock );

        ////////////////////////////////// Wind

        // Speed

        nRawSpeed = htoi( &szLine[WSPD_OFF], RL4 );

        if( nRawSpeed )
        {
            g_nRawWindSpeedAccum += nRawSpeed;
            g_nWindCount++;

            // Direction

            memset( &szLine[WDIR_OFF], '0', 2 );    // per Peet Brothers - force MS word to '00'
            nRawDirection = htoi( &szLine[WDIR_OFF], RL4 );
            g_nRawWindDirAccum += nRawDirection;
            // need to track quadrants to correct for QI/QIV ambiguity
            if( nRawDirection < U2K_90 )
                g_nQ1++;
            else if ( nRawDirection > U2K_270 )
                g_nQ4++;

            // gust

            if( nRawSpeed >= g_nRawGustSpeed )
            {
                g_nRawGustSpeed = nRawSpeed;
                g_nRawGustDirection = nRawDirection;
            }
        }

        ////////////////////////////////// Temperature

        // Outside

        nVal = htoi( &szLine[OTMP_OFF], RL4 );
        nTemp = FixValue( nVal );
        //
        if( nTemp > -50 && nTemp < 150 )
        {
            g_nTemp = nTemp;

            if( g_nTemp > g_nTempHi )
                g_nTempHi = g_nTemp;

            if( g_nTemp < g_nTempLo )
                g_nTempLo = g_nTemp;
        }
        // else ignore

        // Inside

        if( g_Sensors.bInsideTemp )
            g_nInsideTemp = FixValue( htoi( &szLine[ITMP_OFF], RL4 ) );


        ////////////////////////////////// Humidity

        if( g_Sensors.bOutsideHumid )
        {
            nVal = htoi( &szLine[OHUM_OFF], RL4 );
            g_nOutHumid = FixValue( nVal );
        }

        if( g_Sensors.bInsideHumid )
        {
            nVal = htoi( &szLine[IHUM_OFF], RL4 );
            g_nInHumid = FixValue( nVal );
        }

        ////////////////////////////////// Barometer

        if( g_Sensors.bBarom )
            g_nBarom = htoi( &szLine[BARM_OFF], RL4 );

        ////////////////////////////////// Precipitation

        if( g_Sensors.bRainGauge )
        {
            g_nPrecip = htoi( &szLine[RAIN_OFF], RL4 );
            if( g_HostConfig.nGaugeType == TENTHS )
                g_nPrecip *= 10;             // 0.1" gauge
        }

        //////////////////////////////////

        g_nReadCount++;

        pthread_mutex_unlock( &mx_UmxLock );

    }

    return NULL;

}


//////////////////////////////////////////////////////////////////////////////
// Um2Line - retrieves a complete line from the Ultimeter II series
// stations. Converts raw values and stores.  Main service routine must
// perform conversions, if required.
//

void * Um2Line( void * pParam )
{

    char szLine[STRING_SIZE+1];
    int nRawSpeed;
    int nRawDirection;
    void * pBlackHole;

    pBlackHole = pParam;

    //FOREVER
    for(;;)
    {

        ////////////////////////////////// get line from station

        //FOREVER
        for(;;)
        {

            if( g_bStopThread )
                pthread_exit( NULL );

            if( !ReadWxStn( szLine, STRING_SIZE ) )
                continue;

            if( strlen( szLine ) != UM2CNT )
                continue;

            if( (*szLine != '#') && (*szLine != '*') )
                continue;
            else
                break;      // legit line

        }

        pthread_mutex_lock( &mx_UmxLock );

        ////////////////////////////////// Wind

        nRawSpeed = htoi( &szLine[UM2_SPEED], RL2 );

        if( nRawSpeed )
        {
            // Speed
            g_nRawWindSpeedAccum += nRawSpeed;
            g_nWindCount++;

            // Direction

            nRawDirection = htoi( &szLine[UM2_DIR], RL1 );
            g_nRawWindDirAccum += nRawDirection;
            // need to track quadrants to correct for QI/QIV ambiguity
            if( nRawDirection < UM2_90 )
                g_nQ1++;
            else if ( nRawDirection > UM2_270 )
                g_nQ4++;

            // high/low gust

            if( nRawSpeed >= g_nRawGustSpeed )
            {
                g_nRawGustSpeed = nRawSpeed;
                g_nRawGustDirection = nRawDirection;
            }

        }

        ////////////////////////////////// Temperature

        // Outside

        g_nTemp = htoi( &szLine[UM2_TEMP], RL2 ) - 56;

        if( g_nTemp >= g_nTempHi )
            g_nTempHi = g_nTemp;

        if( g_nTemp <= g_nTempLo )
            g_nTempLo = g_nTemp;


        ////////////////////////////////// Precipitation

        if( g_Sensors.bRainGauge )
        {
            g_nPrecip = htoi( &szLine[UM2_PRECIP], RL4 );
            if( g_HostConfig.nGaugeType == TENTHS )
                g_nPrecip *= 10;             // 0.1" gauge
        }


        //////////////////////////////////

        g_nReadCount++;

        pthread_mutex_unlock( &mx_UmxLock );

    }

    return NULL;
}

