/***************************************************************************
 *   Copyright (C) 2005 by John Bennett   *
 *   n4xi@arrl.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

//
// wxport.cpp - serial port handler for weather station
//

#define _WXPORT_CPP_


#include <stdio.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/poll.h>

// libwxn
#include <common.h>
#include <logmsg.h>

#include "wxport.h"


int g_hWxPort = 0;
time_t g_tTimeout;
struct termios g_OldWxIos;
struct termios g_NewWxIos;


BOOL OpenWxStn( int nBaudrate, char * szPort, int nTimeout )
{

    char szDev[32];
    int nBaud;

    g_tTimeout = (time_t) nTimeout * 1000;

    sprintf( szDev, "/dev/%s", szPort );
    g_hWxPort = open( szDev, O_RDWR | O_NOCTTY | O_NONBLOCK );
    if( g_hWxPort < 0 )
    {
        g_hWxPort = 0;
        g_log.PostMsg( HOSTMSG_OPEN_OW_PORT_FAIL, szDev );
        return FALSE;
    }

    // save old termios
    tcgetattr( g_hWxPort, &g_OldWxIos );

    switch( nBaudrate )
    {
        case 1200:
            nBaud = B1200;
            break;

        case 2400:
            nBaud = B2400;
            break;

        case 4800:
            nBaud = B4800;
            break;

        case 9600:
            nBaud = B9600;
            break;

        case 19200:
            nBaud = B19200;
            break;

        default:
            g_log.PostMsg( HOSTMSG_ILLEAGAL_BAUD_RATE, NULL );
            close( g_hWxPort );
            return FALSE;
    }

    bzero( &g_NewWxIos, sizeof( g_NewWxIos ) );
    g_NewWxIos.c_cflag = nBaud | CLOCAL | CREAD | CS8;

    g_NewWxIos.c_iflag = IGNPAR | IGNBRK;
    tcflush( g_hWxPort, TCIFLUSH );
    tcsetattr( g_hWxPort, TCSANOW, &g_NewWxIos );

    return TRUE;

}


void CloseWxStn( void )
{

    if( g_hWxPort )
    {
        tcsetattr( g_hWxPort, TCSANOW, &g_OldWxIos );
        close( g_hWxPort );

        g_hWxPort = 0;
    }

}


//////////////////////////////////////////////////////////////////////////////
// ReadWxStn - write nLength characters to weather station. Returns TRUE
// success, FALSE if timeout.
//

BOOL WriteWxStn( char * szCmd, int nLength )
{

    ssize_t nBytes;

    nBytes = write( g_hWxPort, szCmd, nLength );

    if( nBytes <= 0 )
        return FALSE;
    else
        return TRUE;

}


//////////////////////////////////////////////////////////////////////////////
// ReadWxStn - read a line from the weather station of nMaxLength. Returns
// TRUE success, FALSE if timeout or nMaxLength exceeded.
//

BOOL ReadWxStn( char * szResponse, int nMaxLength )
{

    ssize_t nBytes;
    int i;
    char szBuffer[4];
    struct pollfd Poll;

    i = 0;
    Poll.fd = g_hWxPort;
    Poll.events = POLLIN | POLLPRI;

    //FOREVER
    for(;;)
    {

        if( poll( &Poll, 1, g_tTimeout ) == 0 )
            break;  // return FALSE
        if( (Poll.revents | POLLIN) || (Poll.revents | POLLPRI) )
        {
            nBytes = read( g_hWxPort, szBuffer, 1 );
            if( nBytes == 1 )
            {
                if( szBuffer[0] == CR || szBuffer[0] == 0 )
                    continue;
                else if( szBuffer[0] == LF )
                {
                    szResponse[i] = 0;
                    return TRUE;
                }
                //else
                szResponse[i++] = szBuffer[0];
                if( i == nMaxLength )
                {
                    g_log.PostMsg( HOSTMSG_READWXSTN_BUFFER_OVERFLOW, NULL );
                    return FALSE;
                }
            }
        }

    }

    g_log.PostMsg( HOSTMSG_READWXSTN_TIMEOUT, NULL );

    return FALSE;

}


//////////////////////////////////////////////////////////////////////////////
// ReadWxStnBytes - read n bytes from weather station. Returns TRUE success,
// FALSE if timeout. If timeout occurs, ucResponse returns unchanged.
//

BOOL ReadWxStnBytes( unsigned char * ucResponse, int nLength )
{

    ssize_t nBytes;
    int i;
    char szBuffer[4];
    struct pollfd Poll;

    i = 0;
    Poll.fd = g_hWxPort;
    Poll.events = POLLIN | POLLPRI;

    while( i < nLength )
    {
        if( poll( &Poll, 1, g_tTimeout ) == 0 )
            break;  // return FALSE
        if( (Poll.revents | POLLIN) || (Poll.revents | POLLPRI) )
        {
            nBytes = read( g_hWxPort, szBuffer, 1 );
            if( nBytes == 1 )
            {
                ucResponse[i++] = szBuffer[0];
                if( i == nLength )
                    return TRUE;
            }
        }
    }

    return FALSE;

}


//////////////////////////////////////////////////////////////////////////////
// FlushWxStnBuffer
//

void FlushWxStnBuffer( void )

{

    tcflush( g_hWxPort, TCIFLUSH );

}


//////////////////////////////////////////////////////////////////////////////
// WxStnDataReady - returns TRUE if there is data to be read, FALSE if not.
// Releases 50 mS time-slot while waiting.
//

BOOL WxStnDataReady( void )

{
    struct pollfd Poll;

    Poll.fd = g_hWxPort;
    Poll.events = POLLIN | POLLPRI;

    // wait = 50 mS
    if( poll( &Poll, 1, 50 ) == 0 )
        return FALSE;
    else
        return TRUE;

}


//////////////////////////////////////////////////////////////////////////////
// DrainWxStnBuffer - reads available characters up to nMaxLen until timeout
// or nMaxLen characters exceeded. Returns number of bytes read, or 0 if
// nMaxLen exceeded.
//

int DrainWxStnBuffer( unsigned char * szBuffer, int nMaxLen )

{
    struct pollfd Poll;
    char c;
    int i;

    i = 0;
    Poll.fd = g_hWxPort;
    Poll.events = POLLIN | POLLPRI;

    //FOREVER
    for(;;)
    {
        if( poll( &Poll, 1, 5000 ) == 0 )
            break;  // return

        if( (Poll.revents | POLLIN) || (Poll.revents | POLLPRI) )
        {
            if( read( g_hWxPort, &c, 1 ) )
            {
                *(szBuffer+i) = c;
                i++;
                if( i == nMaxLen )
                    return 0;
            }
        }
    }

    return i;

}

